"""Tests related to theme's patched graphviz ext."""
import pytest
from sphinx.testing.util import SphinxTestApp

WITH_BRACKETS = """
.. graphviz::

    graph {
        "foo()" [xref=":py:func:`foo()`"]
    }

"""

WITHOUT_BRACKETS = """
.. graphviz::

    graph {
        subgraph cluster_foo {
            xref=":py:func:`foo()`"
            A_node
        }
    }

"""


@pytest.mark.parametrize(
    "graph", [WITH_BRACKETS, WITHOUT_BRACKETS], ids=["with", "without"]
)
def test_square_brackets(immaterial_make_app, graph: str):
    """generate a graph with attributes not enclosed in square brackets"""
    app: SphinxTestApp = immaterial_make_app(
        extra_conf="\n".join(
            [
                'extensions.append("sphinx_immaterial.graphviz")',
                # tests also run on Windows in CI
                "graphviz_ignore_incorrect_font_metrics = True",
            ]
        ),
        files={
            "index.rst": """
The Test
========

A function
----------

.. py:function:: foo()

A graph
-------

{}""".format(
                graph
            )
        },
    )

    app.build()
    assert not app._warning.getvalue()



import docutils.nodes


def test_python_class_synopsis(immaterial_make_app, snapshot):

    app = immaterial_make_app(
        files={
            "index.rst": """
.. py:class:: Bar

   Synopsis goes here.

   Rest of description goes here.

:py:obj:`Bar`
"""
        },
    )

    app.build()

    assert not app._warning.getvalue()

    doc = app.env.get_and_resolve_doctree("index", app.builder)

    node = list(doc.findall(condition=docutils.nodes.reference))[-1]

    snapshot.assert_match(node["reftitle"], "reftitle.txt")



from typing import Type, Union, List, cast

import docutils.nodes
import sphinx.domains.python


def _monkey_patch_python_get_signature_prefix(
    directive_cls: Type[sphinx.domains.python.PyObject],
) -> None:
    orig_get_signature_prefix = directive_cls.get_signature_prefix

    def get_signature_prefix(self, sig: str) -> Union[str, List[docutils.nodes.Node]]:
        prefix = orig_get_signature_prefix(self, sig)
        if not self.env.config.python_strip_property_prefix:
            return prefix
        if sphinx.version_info >= (4, 3):
            prefix = cast(List[docutils.nodes.Node], prefix)
            assert isinstance(prefix, list)
            for prop_idx, node in enumerate(prefix):
                if node == "property":
                    assert isinstance(
                        prefix[prop_idx + 1], sphinx.addnodes.desc_sig_space
                    )
                    prefix = list(prefix)
                    del prefix[prop_idx : prop_idx + 2]
                    break
            return prefix
        prefix = cast(str, prefix)  # type: ignore
        assert isinstance(prefix, str)
        parts = prefix.strip().split(" ")
        if "property" in parts:
            parts.remove("property")
        if parts:
            return " ".join(parts) + " "
        return ""

    directive_cls.get_signature_prefix = get_signature_prefix  # type: ignore


_monkey_patch_python_get_signature_prefix(sphinx.domains.python.PyFunction)
_monkey_patch_python_get_signature_prefix(sphinx.domains.python.PyMethod)
_monkey_patch_python_get_signature_prefix(sphinx.domains.python.PyProperty)


def setup(app: sphinx.application.Sphinx):
    app.add_config_value(
        "python_strip_property_prefix", default=False, rebuild="env", types=(bool,)
    )
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# add docs path to python sys.path to allow autodoc-ing a test_py_module
import os
import sys
import typing

from typing_extensions import Literal

sys.path.insert(0, os.path.abspath("."))

import docutils
import sphinx
import sphinx.domains.python
import sphinx.environment
import sphinx.util.logging
import sphinx.util.typing

from sphinx_immaterial.apidoc import (
    object_description_options as _object_description_options,
)

logger = sphinx.util.logging.getLogger(__name__)

# -- Project information -----------------------------------------------------

project = "Sphinx-Immaterial"
copyright = "2021 The Sphinx-Immaterial Authors"
author = "Jeremy Maitin-Shepard"

# The full version, including alpha/beta/rc tags
release = "1"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinxcontrib.details.directive",
    "sphinx_immaterial.theme_result",
    "sphinx_immaterial.kbd_keys",
    "sphinx_immaterial.apidoc.format_signatures",
    "sphinx_immaterial.apidoc.cpp.cppreference",
    "sphinx_immaterial.apidoc.json.domain",
    "sphinx_immaterial.apidoc.python.apigen",
    "sphinx_immaterial.apidoc.cpp.apigen",
    "sphinx_immaterial.graphviz",
    "sphinx_jinja",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "sphinx_docs": ("https://www.sphinx-doc.org/en/master", None),
}

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = "any"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- sphinx_immaterial.keys extension options
#
# optional key_map for example purposes
keys_map = {"my-special-key": "Awesome Key", "git": ""}

# -- Options for HTML output -------------------------------------------------

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named 'default.css' will overwrite the builtin 'default.css'.
html_static_path = ["_static"]
html_css_files = ["extra_css.css"]
html_last_updated_fmt = ""
html_title = "Sphinx-Immaterial"
html_favicon = "_static/images/favicon.ico"  # colored version of material/bookshelf.svg
html_logo = "_static/images/Ybin.gif"  # from https://gifer.com/en/Ybin

# -- HTML theme specific settings ------------------------------------------------

extensions.append("sphinx_immaterial")
html_theme = "sphinx_immaterial"

# material theme options (see theme.conf for more information)
html_theme_options = {
    "icon": {
        "repo": "fontawesome/brands/github",
        "edit": "material/file-edit-outline",
    },
    "site_url": "https://jbms.github.io/sphinx-immaterial/",
    "repo_url": "https://github.com/jbms/sphinx-immaterial/",
    "repo_name": "Sphinx-Immaterial",
    "repo_type": "github",
    "edit_uri": "blob/main/docs",
    "globaltoc_collapse": True,
    "features": [
        "navigation.expand",
        # "navigation.tabs",
        # "toc.integrate",
        "navigation.sections",
        # "navigation.instant",
        # "header.autohide",
        "navigation.top",
        # "navigation.tracking",
        # "search.highlight",
        "search.share",
        "toc.follow",
        "toc.sticky",
        "content.tabs.link",
        "announce.dismiss",
    ],
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "light-green",
            "accent": "light-blue",
            "toggle": {
                "icon": "material/lightbulb-outline",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "deep-orange",
            "accent": "lime",
            "toggle": {
                "icon": "material/lightbulb",
                "name": "Switch to light mode",
            },
        },
    ],
    # BEGIN: version_dropdown
    "version_dropdown": True,
    "version_info": [
        {
            "version": "https://sphinx-immaterial.rtfd.io",
            "title": "ReadTheDocs",
            "aliases": [],
        },
        {
            "version": "https://jbms.github.io/sphinx-immaterial",
            "title": "Github Pages",
            "aliases": [],
        },
    ],
    # END: version_dropdown
    "toc_title_is_page_title": True,
    # BEGIN: social icons
    "social": [
        {
            "icon": "fontawesome/brands/github",
            "link": "https://github.com/jbms/sphinx-immaterial",
        },
        {
            "icon": "fontawesome/brands/python",
            "link": "https://pypi.org/project/sphinx-immaterial/",
        },
    ],
    # END: social icons
}
# end html_theme_options

# ---- Other documentation options -------------------------

todo_include_todos = True

extlinks = {
    "duref": (
        "http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#%s",
        "rST %s",
    ),
    "durole": (
        "http://docutils.sourceforge.net/docs/ref/rst/roles.html#%s",
        "rST role %s",
    ),
    "dudir": (
        "http://docutils.sourceforge.net/docs/ref/rst/directives.html#%s",
        "rST directive %s",
    ),
    "graphvizattr": (
        "https://graphviz.org/docs/attrs/%s/",
        "%s attribute",
    ),
    "dutree": (
        "https://docutils.sourceforge.io/docs/ref/doctree.html#%s",
        "%s",
    ),
}

object_description_options = []

# BEGIN: sphinx_immaterial.apidoc.format_signatures extension options
object_description_options.append(
    ("cpp:.*", dict(clang_format_style={"BasedOnStyle": "LLVM"}))
)
# END: sphinx_immaterial.apidoc.format_signatures extension options

object_description_options.append(("py:.*", dict(wrap_signatures_with_css=True)))

# BEGIN: sphinx_immaterial.apidoc.cpp.external_cpp_references extension options
external_cpp_references = {
    "nlohmann::json": {
        "url": "https://json.nlohmann.me/api/json/",
        "object_type": "type alias",
        "desc": "C++ type alias",
    },
    "nlohmann::basic_json": {
        "url": "https://json.nlohmann.me/api/basic_json/",
        "object_type": "class",
        "desc": "C++ class",
    },
}
# END: sphinx_immaterial.apidoc.cpp.external_cpp_references extension options

# BEGIN: cpp_strip_namespaces_from_signatures option
cpp_strip_namespaces_from_signatures = [
    "my_ns1",
    "my_ns2",
    "my_ns2::my_nested_ns",
]
# END: cpp_strip_namespaces_from_signatures option

rst_prolog = """
.. role:: python(code)
   :language: python
   :class: highlight

.. role:: cpp(code)
   :language: cpp
   :class: highlight

.. role:: json(code)
   :language: json
   :class: highlight

.. role:: rst(code)
   :language: rst
   :class: highlight

.. role:: css(code)
   :language: css
   :class: highlight

.. role:: dot(code)
   :language: dot
   :class: highlight

.. role:: html(code)
   :language: html
   :class: highlight
"""


object_description_options.append(
    (
        "std:confval",
        dict(
            toc_icon_class="data", toc_icon_text="C", generate_synopses="first_sentence"
        ),
    )
)

object_description_options.append(
    (
        "std:objconf",
        dict(
            toc_icon_class="data",
            toc_icon_text="O",
            generate_synopses=None,
        ),
    )
)

object_description_options.append(
    (
        "std:themeconf",
        dict(
            toc_icon_class="data", toc_icon_text="T", generate_synopses="first_sentence"
        ),
    )
)

python_type_aliases = {}

# BEGIN: python_type_aliases example
python_type_aliases = {
    "MyUnqualifiedType": "alias_ex.MyUnqualifiedType",
    "example_mod._internal.": "example_mod.",
}
# END: python_type_aliases example

# BEGIN: python_module_names_to_strip_from_xrefs example
python_module_names_to_strip_from_xrefs = ["tensorstore_demo"]
# END: python_module_names_to_strip_from_xrefs example


jinja_contexts = {
    "sys": {"sys": sys},
}


json_schemas = [
    "apidoc/json/index_transform_schema.yml",
    "apidoc/json/inheritance_schema.yml",
]

json_schema_rst_prolog = """
.. default-role:: json:schema

.. default-literal-role:: json

.. highlight:: json
"""

python_apigen_modules = {"tensorstore_demo": "python_apigen_generated/"}

python_apigen_default_groups = [
    ("class:.*", "Classes"),
    (r".*:.*\.__(init|new)__", "Constructors"),
    (r".*:.*\.__eq__", "Comparison operators"),
    (r".*:.*\.__(str|repr)__", "String representation"),
]

python_apigen_rst_prolog = """
.. default-role:: py:obj

.. default-literal-role:: python

.. highlight:: python
"""

cpp_demo_include_dir = os.path.join(os.path.dirname(__file__))

cpp_apigen_configs = [
    dict(
        document_prefix="cpp_apigen_generated/",
        api_parser_config=dict(
            input_content="""
#include "cpp_apigen_demo/index_interval.h"
#include "cpp_apigen_demo/array.h"
""",
            compiler_flags=["-std=c++17", "-I", cpp_demo_include_dir, "-x", "c++"],
            include_directory_map={
                f"{cpp_demo_include_dir}/": "",
            },
            allow_paths=["^cpp_apigen_demo/"],
            disallow_namespaces=["^std$"],
            verbose=True,
        ),
    ),
]

cpp_apigen_rst_prolog = """
.. default-role:: cpp:expr

.. default-literal-role:: cpp

.. highlight:: cpp
"""

autodoc_class_signature = "separated"

nitpicky = True
nitpick_ignore = [
    # Python standard library types not present in object inventory.
    ("py:class", "Pattern"),
    ("py:class", "re.Pattern"),
    # Config option type
    ("py:class", "ExternalCppReference"),
    # Example Python types
    ("py:class", "example_mod.Foo"),
    ("py:class", "alias_ex.MyUnqualifiedType"),
    # Example C++ types
    ("cpp:identifier", "Sphinx"),
    ("cpp:identifier", "RF24_SPI_SPEED"),
    ("cpp:identifier", "RF24_SPI_SPEED"),
    # C++ namespaces referenced in the documentation
    #
    # It is a bug in the C++ domain that a reference to `ns::symbol` will
    # ultimately generate a reference to both `ns` and `ns::symbol`.  However,
    # because the C++ domain does not actually define "namespace" objects, the
    # `ns` reference will always fail to be resolved, leading to a spurious
    # warning.
    ("cpp:identifier", "::nlohmann"),
    ("cpp:identifier", "std"),
    ("cpp:identifier", "synopses_ex"),
    ("cpp:identifier", "my_ns1"),
    ("cpp:identifier", "my_ns2"),
    ("cpp:identifier", "my_ns2::my_nested_ns"),
    ("cpp:identifier", "my_ns3"),
    ("cpp:identifier", "cpp_apigen_demo"),
    ("cpp:identifier", "::nlohmann"),
    ("cpp:identifier", "std"),
    ("cpp:identifier", "synopses_ex"),
    ("cpp:identifier", "my_ns1"),
    ("cpp:identifier", "my_ns2"),
    ("cpp:identifier", "my_ns2::my_nested_ns"),
    ("cpp:identifier", "my_ns3"),
    # Example JavaScript types
    ("js:func", "string"),
    ("js:func", "SomeError"),
]


graphviz_ignore_incorrect_font_metrics = True


def _validate_parallel_build(app):
    # Verifies that all of the extensions defined by this theme support parallel
    # building.
    assert app.is_parallel_allowed("read")
    assert app.is_parallel_allowed("write")


def _parse_object_description_signature(
    env: sphinx.environment.BuildEnvironment, signature: str, node: docutils.nodes.Node
) -> str:
    registry = _object_description_options.get_object_description_option_registry(
        env.app
    )
    registry_option = registry.get(signature)
    node += sphinx.addnodes.desc_name(signature, signature)
    if registry_option is None:
        logger.error("Invalid object description option: %r", signature, location=node)
    else:
        node += sphinx.addnodes.desc_sig_punctuation(" : ", " : ")
        annotations = sphinx.domains.python._parse_annotation(
            sphinx.util.typing.stringify(registry_option.type_constraint), env
        )
        node += sphinx.addnodes.desc_type("", "", *annotations)
        node += sphinx.addnodes.desc_sig_punctuation(" = ", " = ")
        default_repr = repr(registry_option.default)
        node += docutils.nodes.literal(
            default_repr,
            default_repr,
            language="python",
            classes=["python", "code", "highlight"],
        )
    return signature


def _parse_confval_signature(
    env: sphinx.environment.BuildEnvironment, signature: str, node: docutils.nodes.Node
) -> str:
    values = env.config.values
    registry_option = values.get(signature)
    node += sphinx.addnodes.desc_name(signature, signature)
    if registry_option is None:
        logger.error("Invalid config option: %r", signature, location=node)
    else:
        default, rebuild, types = registry_option
        if isinstance(types, sphinx.config.ENUM):
            types = (Literal[tuple(types.candidates)],)
        if isinstance(types, type):
            types = (types,)
        if types:
            type_constraint = typing.Union[tuple(types)]
            node += sphinx.addnodes.desc_sig_punctuation(" : ", " : ")
            annotations = sphinx.domains.python._parse_annotation(
                sphinx.util.typing.stringify(type_constraint), env
            )
            node += sphinx.addnodes.desc_type("", "", *annotations)
        if not callable(default):
            node += sphinx.addnodes.desc_sig_punctuation(" = ", " = ")
            default_repr = repr(default)
            node += docutils.nodes.literal(
                default_repr,
                default_repr,
                language="python",
                classes=["python", "code", "highlight"],
            )
    return signature


def setup(app):
    app.add_object_type(
        "confval",
        "confval",
        objname="configuration value",
        indextemplate="pair: %s; configuration value",
        parse_node=_parse_confval_signature,
    )

    app.add_object_type(
        "themeconf",
        "themeconf",
        objname="theme configuration option",
        indextemplate="pair: %s; theme option",
    )

    app.add_object_type(
        "objconf",
        "objconf",
        objname="object description option",
        indextemplate="pair: %s; object description option",
        parse_node=_parse_object_description_signature,
    )

    # Add `event` type from Sphinx's own documentation, to allow intersphinx
    # references to Sphinx events.
    app.add_object_type(
        "event",
        "event",
        objname="Sphinx event",
        indextemplate="pair: %s; event",
    )
    app.connect("builder-inited", _validate_parallel_build)



"""C++ API documentation generation extension."""

import collections
import contextlib
import dataclasses
import json
import os
import re
from typing import cast, Dict, List, Any, Optional, Tuple, Set, Type

import docutils.core
import docutils.nodes
import docutils.parsers.rst.directives
import docutils.utils
import pydantic
import pydantic.dataclasses
import sphinx.application
import sphinx.builders
import sphinx.builders.html
import sphinx.domains.c
import sphinx.domains.cpp
import sphinx.environment
import sphinx.errors
import sphinx.util.cfamily
import sphinx.util.docfields
import sphinx.util.logging


from .. import apigen_utils
from ... import sphinx_utils
from . import api_parser
from ... import default_literal_role
from .signodes import desc_cpp_explicit

logger = sphinx.util.logging.getLogger(__name__)

APIDOC_SECTION_ID_PREFIX = "cpp-api-"


@pydantic.dataclasses.dataclass
class ApigenConfig:
    """Specifies a C++ API generation configuration."""

    document_prefix: str
    """Sphinx document path prefix for the per-entity documentation pages.

    .. important::
        This should use the Linux path separator :python:`"/"` (even if
        building on Windows).
    """

    api_parser_config: Optional[api_parser.Config] = None
    """Configuration for generating an API description.

    This option and `.api_data` are mutually exclusive.
    """

    api_data: Optional[str] = None
    """Path to already-parsed API description.

    This option and `.api_parser_config` are mutually exclusive.
    """


CppApiEntity = Any
EntityId = str


@dataclasses.dataclass
class CppApiData:
    groups: Dict[str, List[EntityId]] = dataclasses.field(default_factory=dict)
    entities: Dict[EntityId, CppApiEntity] = dataclasses.field(default_factory=dict)
    nonitpick: Set[Tuple[str, str, int]] = dataclasses.field(default_factory=set)

    def get_entity_sub_groups(
        self, entity_id: str, member: bool
    ) -> Dict[str, List[EntityId]]:
        entity = self.entities[entity_id]
        return entity.get("related_members" if member else "related_nonmembers", {})

    def get_documented_entities(self):
        seen = set()

        def process_group(entity_ids: List[str]):
            for entity_id in entity_ids:
                if entity_id in seen:
                    continue
                seen.add(entity_id)
                yield entity_id
                for member in (True, False):
                    for related_ids in self.get_entity_sub_groups(
                        entity_id, member=member
                    ).values():
                        yield from process_group(related_ids)

        for group_members in self.groups.values():
            yield from process_group(group_members)

    def get_entity_toc_title(self, entity: CppApiEntity) -> str:
        return entity["name"]

    def get_entity_scope(self, entity: CppApiEntity) -> str:
        components = []
        cur_entity = entity
        while True:
            parent_id = cur_entity.get("parent")
            if parent_id is None:
                break
            parent_entity = self.entities.get(parent_id)
            assert parent_entity is not None
            cur_entity = parent_entity
            name_with_args = cur_entity["name"]
            if not cur_entity.get("specializes"):
                name_with_args += _format_template_arguments(cur_entity)
            components.append(name_with_args)
        components.reverse()
        if components:
            components.append("")
        return cur_entity.get("scope", "") + "::".join(components)

    def get_entity_object_name(self, entity: CppApiEntity) -> str:
        name = self.get_entity_scope(entity) + entity["name"]
        special_id = entity.get("special_id")
        if special_id:
            name += f"[{special_id}]"
        return name

    def get_entity_page_name(self, entity: CppApiEntity) -> str:
        return entity["page_name"]

    def get_entity_page_path(self, entity: CppApiEntity) -> str:
        return entity["document_prefix"] + self.get_entity_page_name(entity)


def _get_cpp_api_data(
    env: sphinx.environment.BuildEnvironment, warn: bool = False
) -> CppApiData:
    KEY = "_sphinx_immaterial_cpp_apigen_data"
    data = getattr(env, KEY, None)
    if data is not None:
        return data

    apigen_configs = getattr(env.app, _CONFIG_ATTR)
    data = CppApiData()

    for apigen_config in apigen_configs:
        if apigen_config.api_parser_config is not None:
            json_data = api_parser.generate_output(apigen_config.api_parser_config)
        else:
            with open(
                os.path.join(env.app.srcdir, apigen_config.api_data),
                "r",
                encoding="utf-8",
            ) as f:
                json_data = json.load(f)
        for entity in json_data["entities"].values():
            entity["document_prefix"] = apigen_config.document_prefix
        data.entities.update(json_data["entities"])
        for group_id, group in json_data["groups"].items():
            data.groups.setdefault(group_id, []).extend(group)
        data.nonitpick.update(
            set((x["target"], x["file"], x["line"] + 1) for x in json_data["nonitpick"])
        )

        if warn:
            for msg, location in json_data["warnings"]:
                if location is not None:
                    location = tuple(location)
                logger.warning(msg, location=location)
            for msg, location in json_data["errors"]:
                if location is not None:
                    location = tuple(location)
                logger.error(msg, location=location)

    setattr(env, KEY, data)
    return data


def _transform_doc_comment(
    env: sphinx.environment.BuildEnvironment, doc, summary: bool
) -> Optional[docutils.statemachine.StringList]:
    if not doc:
        return None
    text = doc["text"]
    location = doc["location"]
    filename = location["file"]
    lineno = location["line"]

    if summary:
        # Find first blank line
        first_blank_line = re.search(r"^\s*$", text, flags=re.MULTILINE)
        if first_blank_line:
            text = text[: first_blank_line.start()]

    out = docutils.statemachine.StringList()
    sphinx_utils.append_multiline_string_to_stringlist(
        out,
        ".. highlight-push::\n\n" + env.config.cpp_apigen_rst_prolog + "\n\n",
        "<cpp_apigen_rst_prolog>",
        -2,
    )
    sphinx_utils.append_multiline_string_to_stringlist(out, text, filename, lineno)
    sphinx_utils.append_multiline_string_to_stringlist(
        out,
        "\n\n" + env.config.cpp_apigen_rst_epilog + "\n\n.. highlight-pop::\n\n",
        "<cpp_apigen_rst_epilog>",
        -2,
    )
    return out


def _format_template_parameters(entity: CppApiEntity) -> str:
    template_parameters = entity.get("template_parameters")
    if template_parameters is None:
        return ""
    tparam_list_str = ", ".join(x["declaration"] for x in template_parameters)
    tparams = f"template <{tparam_list_str}> "
    requires = entity.get("requires")
    if requires:
        requires_expr = " && ".join(requires)
        tparams += f"requires {requires_expr} "
    return tparams


def _format_template_arguments(entity: CppApiEntity) -> str:
    if entity.get("specializes"):
        # Template arguments already included in `entity["name"]`.
        return ""
    template_parameters = entity.get("template_parameters")
    if not template_parameters:
        return ""
    strs = []
    for param in template_parameters:
        arg = param["name"]
        if not arg:
            continue
        if param["pack"]:
            arg += "..."
        strs.append(arg)
    args_str = ", ".join(strs)
    return f"<{args_str}>"


def _validate_signature_line_tags(
    strip_tags: List[Type[sphinx.addnodes.desc_sig_element]],
    children: List[docutils.nodes.Node],
):
    for strip_tag, child in zip(strip_tags, children):
        if isinstance(child, sphinx.addnodes.desc_signature_line) and child.children:
            child = child.children[0]
        if not isinstance(child, strip_tag):
            raise ValueError(
                f"Expected {strip_tag!r} tag, but received: {type(child)!r}"
            )


def _strip_template_parameters(
    api_data: CppApiData,
    entity: CppApiEntity,
    signode: sphinx.addnodes.desc_signature,
    strip_self_parameters: bool = False,
    strip_ancestors_parameters: bool = False,
):
    cur_entity = entity
    strip_leading: List[str] = []
    strip_trailing: List[str] = []
    if strip_self_parameters:
        if entity.get("requires") and entity.get("template_parameters") is None:
            strip_trailing.append("trailingRequiresClause")
    depth = 0
    while True:
        if depth > 0 and not strip_ancestors_parameters:
            break
        if depth > 0 or strip_self_parameters:
            if cur_entity.get("template_parameters") is not None:
                strip_leading.append("templateParams")
                if cur_entity.get("requires"):
                    strip_leading.append("requiresClause")
        depth += 1
        parent_id = cur_entity.get("parent")
        if parent_id is None:
            break
        cur_entity = api_data.entities[parent_id]

    signature_lines: List[Tuple[int, sphinx.addnodes.desc_signature_line]] = [
        (i, child)
        for i, child in enumerate(signode)
        if isinstance(child, sphinx.addnodes.desc_signature_line)
    ]
    if len(signature_lines) < len(strip_leading) + len(strip_trailing):
        raise ValueError("Missing children")
    for i, line_type in enumerate(strip_leading):
        assert signature_lines[i][1].sphinx_line_type == line_type

    for i, line_type in enumerate(strip_trailing):
        assert (
            signature_lines[len(signature_lines) - 1 - i][1].sphinx_line_type
            == line_type
        )

    remove = (
        signature_lines[: len(strip_leading)]
        + signature_lines[len(signature_lines) - len(strip_trailing) :]
    )
    remove.reverse()
    for i, _ in remove:
        del signode[i]


def _summarize_explicit(signode: sphinx.addnodes.desc_signature) -> None:
    for node in signode.findall(condition=desc_cpp_explicit):
        if len(node.children) == 1:
            continue
        node.children[2:-1] = sphinx.addnodes.desc_sig_punctuation("...", "...")


def _format_entity_alias(entity: CppApiEntity, full_name: str) -> Tuple[str, str]:
    underlying_type = entity["underlying_type"]
    signature = f"{full_name}"
    if underlying_type is not None:
        signature += f" = {underlying_type}"
    return ("cpp:type", signature)


def _format_entity_constructor(entity: CppApiEntity, full_name: str) -> Tuple[str, str]:
    return _format_entity_function(entity, full_name)


def _format_entity_class(entity: CppApiEntity, full_name: str) -> Tuple[str, str]:
    prefix = " ".join(entity["prefix"])
    signature = f"{prefix} {full_name}"
    base_strs = []
    for base in entity["bases"]:
        base_strs.append(f'{base["access"]} {base["type"]}')
    if base_strs:
        signature += " : "
        signature += ", ".join(base_strs)

    return (f'cpp:{entity["keyword"]}', signature)


def _format_entity_function(entity: CppApiEntity, full_name: str) -> Tuple[str, str]:
    signature = entity["declaration"].replace(entity["name_substitute"], full_name)
    return ("cpp:function", signature)


def _format_entity_method(entity: CppApiEntity, full_name: str) -> Tuple[str, str]:
    return _format_entity_function(entity, full_name)


def _format_entity_conversion_function(
    entity: CppApiEntity, full_name: str
) -> Tuple[str, str]:
    return _format_entity_function(entity, full_name)


def _format_entity_var(entity: CppApiEntity, full_name: str) -> Tuple[str, str]:
    declaration = entity["declaration"].replace(entity["name_substitute"], full_name)
    signature = declaration
    initializer = entity["initializer"]
    if initializer is not None:
        signature += initializer
    return ("cpp:var", signature)


def _format_entity_enum(entity: CppApiEntity, full_name: str) -> Tuple[str, str]:
    signature = full_name
    directive_name = "cpp:enum"
    if entity["keyword"]:
        directive_name += f'-{entity["keyword"]}'
    return (directive_name, signature)


def _format_entity_macro(entity: CppApiEntity, full_name: str) -> Tuple[str, str]:
    signature = full_name
    parameters = entity["parameters"]
    if parameters is not None:
        signature += f'({", ".join(parameters)})'
    return ("c:macro", signature)


def _format_template_prefix(
    api_data: CppApiData, entity: CppApiEntity, summary: bool
) -> str:
    inner_prefix = _format_template_parameters(entity)
    if summary:
        return inner_prefix
    prefixes = [inner_prefix]
    while True:
        parent_id = entity.get("parent")
        if parent_id is None:
            break
        entity = api_data.entities[parent_id]
        prefixes.append(_format_template_parameters(entity))
    prefixes.reverse()
    return " ".join(prefixes)


def _format_trailing_requires(entity: CppApiEntity, summary: bool) -> str:
    template_parameters = entity.get("template_parameters")
    requires = entity.get("requires")
    if template_parameters is not None or not requires:
        return ""
    return " requires " + " && ".join(expr for expr in requires)


def _format_entity(
    api_data: CppApiData, entity: CppApiEntity, summary: bool, include_scope: bool
):
    full_name = api_data.get_entity_scope(entity) if include_scope else ""
    full_name += entity["name"]
    kind = entity["kind"]
    if summary and kind in ("class", "alias", "var"):
        # Add template arguments
        full_name += _format_template_arguments(entity)
    cpp_directive, signature_text = globals()[f"_format_entity_{kind}"](
        entity, full_name
    )
    signature_line = (
        _format_template_prefix(api_data, entity, summary=summary)
        + signature_text
        + _format_trailing_requires(entity, summary=summary)
    )
    if kind != "macro":
        signature_line += ";"
    return cpp_directive, signature_line


@contextlib.contextmanager
def save_cpp_scope(env: sphinx.environment.BuildEnvironment):
    parent_symbol = env.temp_data.get("cpp:parent_symbol", False)
    namespace_stack = env.temp_data.get("cpp:namespace_stack", False)
    if namespace_stack:
        namespace_stack = list(namespace_stack)
    parent_key = env.ref_context.get("cpp:parent_key", False)

    yield

    if parent_symbol is not False:
        env.temp_data["cpp:parent_symbol"] = parent_symbol
    else:
        env.temp_data.pop("cpp:parent_symbol", None)

    if namespace_stack is not False:
        env.temp_data["cpp:namespace_stack"] = namespace_stack
    else:
        env.temp_data.pop("cpp:namespace_stack", None)

    if parent_key is not False:
        env.ref_context["cpp:parent_key"] = parent_key
    else:
        env.ref_context.pop("cpp:parent_key", None)


def _add_entity_description(
    env: sphinx.environment.BuildEnvironment,
    api_data: CppApiData,
    entity: CppApiEntity,
    contentnode: docutils.nodes.Element,
    summary: bool,
    state: docutils.parsers.rst.states.RSTState,
) -> None:
    kind = entity["kind"]
    out = docutils.statemachine.StringList()
    location = entity["location"]
    parent_id = entity.get("parent")
    if parent_id and summary:
        include_scope = False
    else:
        include_scope = True

    if parent_id:
        parent = api_data.entities[parent_id]
        scope_template_prefix = _format_template_prefix(
            api_data, api_data.entities[parent_id], summary=False
        )
    else:
        parent = None
        scope_template_prefix = ""

    scope = api_data.get_entity_scope(entity).rstrip(":")

    sphinx_utils.append_directive_to_stringlist(
        out,
        "cpp:namespace",
        scope_template_prefix + scope if not include_scope else "nullptr",
        source_path=location["file"],
        source_line=location["line"],
    )

    cpp_directive, signature_line = _format_entity(
        api_data, entity, summary, include_scope
    )

    signature_lines = []

    num_include_lines = 0

    if not summary:
        include_path = location["file"]
        signature_lines.append(f'#include "{include_path}"')
        num_include_lines = 1

    signature_lines.append(signature_line)
    siblings = entity.get("siblings")
    entity_ids = [entity["id"]]
    if siblings:
        for sibling_id in siblings:
            sibling = api_data.entities[sibling_id]
            other_cpp_directive, other_signature_line = _format_entity(
                api_data, sibling, summary, include_scope
            )
            assert other_cpp_directive == cpp_directive
            signature_lines.append(other_signature_line)
            entity_ids.append(sibling["id"])

    body = _transform_doc_comment(env, entity["doc"], summary=summary)
    sphinx_utils.append_directive_to_stringlist(
        out,
        cpp_directive,
        "\n".join(signature_lines),
        options={
            "noindexentry": True,
            "noindex": summary,
            "node-id": "",
            "symbol-ids": json.dumps(entity_ids),
        },
        source_path=location["file"],
        source_line=location["line"],
        content=body,
    )

    with apigen_utils.save_rst_defaults(env), save_cpp_scope(env):
        nodes = [
            x
            for x in sphinx_utils.parse_rst(state=state, text=out)
            if isinstance(x, sphinx.addnodes.desc)
        ]
    assert len(nodes) == 1
    objdesc = nodes[0]

    desc_name_nodes: List[docutils.nodes.Element] = []

    for signode in objdesc.children[num_include_lines:-1]:
        assert isinstance(signode, sphinx.addnodes.desc_signature)
        # Find the node in the signature containing the actual unqualified name.
        for sub_node in signode.findall(condition=sphinx.addnodes.desc_name):
            if "sig-name-nonprimary" in sub_node["classes"]:
                continue
            desc_name_nodes.append(sub_node)
            break
        else:
            raise ValueError("Failed to find desc_name node")

    obj_content = cast(sphinx.addnodes.desc_content, objdesc.children[-1])

    for entity_id, signode in zip(entity_ids, objdesc.children[num_include_lines:-1]):
        assert isinstance(signode, sphinx.addnodes.desc_signature)
        _strip_template_parameters(
            api_data,
            api_data.entities[entity_id],
            signode,
            strip_ancestors_parameters=not summary,
            strip_self_parameters=summary,
        )
        if summary:
            _summarize_explicit(signode)

    if summary:
        objdesc["classes"].append("summary")
        # Insert a link around the `desc_name` field
        for i, desc_name_node in enumerate(desc_name_nodes):
            xref_node = sphinx.addnodes.pending_xref(
                "",
                desc_name_node.deepcopy(),
                refdomain="std",
                reftype="doc",
                reftarget="/" + api_data.get_entity_page_path(entity),
                refwarn=True,
                refexplicit=True,
            )
            desc_name_node.replace_self(xref_node)
            desc_name_nodes[i] = xref_node
    else:
        if entity["kind"] == "enum":
            _add_enumerators(
                env=env,
                api_data=api_data,
                entity=entity,
                obj_content=obj_content,
                parent_scope=scope_template_prefix + scope,
                state=state,
            )
        member_groups = api_data.get_entity_sub_groups(entity["id"], member=True)
        _merge_summary_nodes_into(
            env=env,
            api_data=api_data,
            state=state,
            contentnode=obj_content,
            groups=member_groups,
            top_level_group=False,
        )

    contentnode += nodes


def _add_enumerators(
    env: sphinx.environment.BuildEnvironment,
    api_data: CppApiData,
    entity: CppApiEntity,
    parent_scope: str,
    obj_content: sphinx.addnodes.desc_content,
    state: docutils.parsers.rst.states.RSTState,
) -> None:
    parent_object_name = api_data.get_entity_object_name(entity)
    for child in entity["enumerators"]:
        name = child["name"]
        location = child["location"]
        out = docutils.statemachine.StringList()
        sphinx_utils.append_directive_to_stringlist(
            out,
            "cpp:namespace",
            f'{parent_scope}::{entity["name"]}',
            source_path=location["file"],
            source_line=location["line"],
        )
        anchor = f"e-{name}"
        sphinx_utils.append_directive_to_stringlist(
            out,
            "cpp:enumerator",
            child["decl"],
            options={
                "noindexentry": True,
                "symbol-ids": json.dumps([child["id"]]),
                "node-id": anchor,
            },
            source_path=location["file"],
            source_line=location["line"],
            content=_transform_doc_comment(env, child["doc"], summary=False),
        )
        with apigen_utils.save_rst_defaults(env), save_cpp_scope(env):
            nodes = [
                x
                for x in sphinx_utils.parse_rst(state=state, text=out)
                if isinstance(x, sphinx.addnodes.desc)
            ]
        assert len(nodes) == 1
        enumerator_objdesc = nodes[0]
        obj_content += enumerator_objdesc
        enumerator_signode, enumerator_doc_node = enumerator_objdesc.children
        assert isinstance(enumerator_signode, sphinx.addnodes.desc_signature)
        enumerator_signode["toc_title"] = name


def _add_group_summary(
    env: sphinx.environment.BuildEnvironment,
    api_data: CppApiData,
    state: docutils.parsers.rst.states.RSTState,
    entities: List[EntityId],
    parent: docutils.nodes.Element,
    notoc: bool = False,
) -> None:
    toc_entries = []

    name_counts = collections.Counter(
        api_data.entities[entity_id]["name"] for entity_id in entities
    )

    for entity_id in entities:
        entity = api_data.entities[entity_id]
        path = api_data.get_entity_page_path(entity)
        _add_entity_description(
            env=env,
            api_data=api_data,
            entity=entity,
            contentnode=parent,
            summary=True,
            state=state,
        )
        title = entity["name"]
        if name_counts[title] > 1:
            special_id = entity.get("special_id")
            if special_id is not None:
                title += f" [{special_id}]"
        if not notoc:
            toc_entries.append((title, path))

    parent.extend(
        sphinx_utils.make_toctree_node(
            state=state,
            toc_entries=toc_entries,
            options={"hidden": True},
            source_path=__name__,
        )
    )


def _merge_summary_nodes_into(
    env: sphinx.environment.BuildEnvironment,
    api_data: CppApiData,
    state: docutils.parsers.rst.states.RSTState,
    contentnode: docutils.nodes.Element,
    groups: Dict[str, List[EntityId]],
    top_level_group: bool,
) -> None:
    """Merges the member summary into `contentnode`.

    Members are organized into groups.  The group is either specified explicitly
    by a `Group:` field in the docstring, or determined automatically by
    `_get_group_name`.  If there is an existing section, the member summary is
    appended to it.  Otherwise, a new section is created.

    Args:
      contentnode: The existing container to which the member summaries will be
        added.  If `contentnode` contains sections, those sections correspond to
        group names.
    """

    apigen_utils.merge_groups_into(
        parent=contentnode,
        group_id_prefix=APIDOC_SECTION_ID_PREFIX if top_level_group else "",
        groups=groups,
        insert_group=lambda entities, section: _add_group_summary(
            env=env, api_data=api_data, state=state, entities=entities, parent=section
        ),
    )


class CppApigenEntityPageDirective(sphinx.util.docutils.SphinxDirective):
    """Documents an entity and summarizes its members, if applicable."""

    has_content = False
    final_argument_whitespace = True
    required_arguments = 1
    optional_arguments = 0

    def run(self) -> List[docutils.nodes.Node]:
        # Wrap in a section
        section = docutils.nodes.section()
        section["ids"].append("")

        # Sphinx treates the first child of a `section` node as the title,
        # regardless of its type.  We use a comment node to avoid adding a title
        # that would be redundant with the object description.
        comment_placeholder = docutils.nodes.comment("", "")
        section += comment_placeholder

        api_data = _get_cpp_api_data(self.env)

        entity_id = self.arguments[0]
        entity = api_data.entities.get(entity_id)
        if entity is None:
            logger.error(
                "Undefined C++ entity: %r", entity_id, location=self.get_location()
            )
            return []

        _add_entity_description(
            env=self.env,
            api_data=api_data,
            entity=entity,
            contentnode=section,
            summary=False,
            state=self.state,
        )
        comment_placeholder.replace_self(
            docutils.nodes.comment("", api_data.get_entity_object_name(entity))
        )

        non_member_groups = api_data.get_entity_sub_groups(entity_id, member=False)
        _merge_summary_nodes_into(
            env=self.env,
            api_data=api_data,
            state=self.state,
            contentnode=section,
            groups=non_member_groups,
            top_level_group=False,
        )
        return [section]


class CppApigenTopLevelGroupDirective(sphinx.util.docutils.SphinxDirective):
    """Summarizes the members of a top-level group."""

    has_content = False
    final_argument_whitespace = True
    required_arguments = 1
    optional_arguments = 0

    option_spec = {
        "notoc": docutils.parsers.rst.directives.flag,
    }

    def run(self) -> List[docutils.nodes.Node]:
        env = self.env
        api_data = _get_cpp_api_data(env)
        group_id = docutils.nodes.make_id(self.arguments[0])
        entities = api_data.groups.get(group_id)
        if not entities:
            logger.warning(
                "No entities in group %r",
                group_id,
                location=self.get_location(),
            )
            return []
        contentnode = docutils.nodes.section()
        _add_group_summary(
            env=self.env,
            api_data=api_data,
            state=self.state,
            entities=entities,
            parent=contentnode,
            notoc="notoc" in self.options,
        )
        return contentnode.children


class CppApigenEntitySummaryDirective(sphinx.util.docutils.SphinxDirective):
    """Generates a summary/link to a C++ entity."""

    has_content = False
    final_argument_whitespace = True
    required_arguments = 1
    optional_arguments = 0

    option_spec = {
        "notoc": docutils.parsers.rst.directives.flag,
    }

    def run(self) -> List[docutils.nodes.Node]:
        env = self.env
        api_data = _get_cpp_api_data(env)
        entity_id = self.arguments[0]
        entity = api_data.entities.get(entity_id)
        if entity is None:
            logger.warning(
                "No C++ entity named: %r",
                entity_id,
                location=self.get_location(),
            )
            return []
        contentnode = docutils.nodes.section()
        _add_group_summary(
            env=self.env,
            api_data=api_data,
            state=self.state,
            entities=[entity_id],
            parent=contentnode,
            notoc="notoc" in self.options,
        )
        return contentnode.children


def _builder_inited(app: sphinx.application.Sphinx) -> None:
    """Generates the rST files for API members."""

    api_data = _get_cpp_api_data(app.env, warn=True)
    apigen_configs = getattr(app, _CONFIG_ATTR)

    writer = apigen_utils.GeneratedDocumentWriter(
        app=app,
        case_insensitive_filesystem=app.config.cpp_apigen_case_insensitive_filesystem,
        output_prefixes=[
            apigen_config.document_prefix for apigen_config in apigen_configs
        ],
        generator_module=__name__,
    )
    writer.prepare_output_directories()
    for entity in api_data.entities.values():
        page_name = entity.get("page_name")
        if page_name is None:
            continue
        entity["page_name"] = apigen_utils.make_unique_docname(
            page_name, writer.case_insensitive_filesystem
        )

    def get_entities():
        for entity_id in api_data.get_documented_entities():
            entity = api_data.entities[entity_id]
            content = sphinx_utils.format_directive(
                "cpp-apigen-entity-page",
                entity_id,
            )
            object_name = api_data.get_entity_object_name(entity)
            docname = api_data.get_entity_page_path(entity)
            yield (docname, object_name, content)

    writer.write_files(get_entities())


def _warn_missing_reference(
    app: sphinx.application.Sphinx, domain, node: sphinx.addnodes.pending_xref
) -> bool:
    if not isinstance(domain, sphinx.domains.cpp.CPPDomain):
        return False
    api_data = _get_cpp_api_data(app.env)
    source, line = docutils.utils.get_source_line(node)
    return (node["reftarget"], source, line) in api_data.nonitpick


_CONFIG_ATTR = "_sphinx_immaterial_cpp_apigen_configs"


def _config_inited(
    app: sphinx.application.Sphinx, config: sphinx.config.Config
) -> None:
    apigen_configs = pydantic.parse_obj_as(
        List[ApigenConfig], config.cpp_apigen_configs
    )
    setattr(app, _CONFIG_ATTR, apigen_configs)


def setup(app: sphinx.application.Sphinx):
    app.connect("config-inited", _config_inited)
    app.connect("builder-inited", _builder_inited)
    app.connect("warn-missing-reference", _warn_missing_reference)
    app.add_directive("cpp-apigen-entity-page", CppApigenEntityPageDirective)
    app.add_directive("cpp-apigen-group", CppApigenTopLevelGroupDirective)
    app.add_directive("cpp-apigen-entity-summary", CppApigenEntitySummaryDirective)
    app.add_config_value(
        "cpp_apigen_configs", default=[], rebuild="env", types=(List[ApigenConfig],)
    )
    app.add_config_value(
        "cpp_apigen_case_insensitive_filesystem",
        default=None,
        types=(bool, type(None)),
        rebuild="env",
    )
    app.add_config_value(
        "cpp_apigen_rst_prolog", types=(str,), default="", rebuild="env"
    )
    app.add_config_value(
        "cpp_apigen_rst_epilog", types=(str,), default="", rebuild="env"
    )
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""A directive designed to reduce example snippets duplication."""
from docutils.parsers.rst import directives
from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from sphinx.directives.code import container_wrapper


class ResultsDirective(SphinxDirective):
    """Takes content as a rst code snippet and renders it after outputting it as a
    literal code-block."""

    has_content = True
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {
        "class": directives.class_option,
        "name": directives.unchanged,
        "output-prefix": directives.unchanged,
    }

    def run(self):
        """Run the directive."""
        container_node = nodes.container(
            "", classes=self.options.get("class", []) + ["results"]
        )
        code = "\n".join(self.content)
        literal_node = nodes.literal_block(code, code)
        literal_node["language"] = "rst"
        if self.arguments:
            literal_node = container_wrapper(self, literal_node, self.arguments[0])
        container_node += literal_node

        if "output-prefix" in self.options:
            out_prefix = nodes.container("", classes=["results-prefix"])
            prefix = self.options.get("output-prefix", "")
            if not prefix:
                prefix = "Which renders the following content:"
            textnodes, _ = self.state.inline_text(prefix, self.lineno)
            out_prefix += textnodes
            container_node += out_prefix

        # the `result` CSS class is theme specific (as used by mkdocs-material theme)
        results_div = nodes.container("", classes=["result"])
        self.state.nested_parse(self.content, self.content_offset, results_div)
        container_node += results_div

        self.set_source_info(container_node)
        self.add_name(container_node)
        return [container_node]


def setup(app):
    app.add_directive("rst-example", ResultsDirective)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from typing import Type, Tuple

import docutils.nodes
import sphinx.addnodes
import sphinx.domains.python


def _monkey_patch_pyattribute_handle_signature(
    directive_cls: Type[sphinx.domains.python.PyObject],
):
    """Modifies PyAttribute or PyVariable to improve styling of signature."""

    def handle_signature(
        self, sig: str, signode: sphinx.addnodes.desc_signature
    ) -> Tuple[str, str]:
        result = super(directive_cls, self).handle_signature(sig, signode)
        typ = self.options.get("type")
        if typ:
            signode += sphinx.addnodes.desc_sig_punctuation("", " : ")
            signode += sphinx.domains.python._parse_annotation(typ, self.env)

        value = self.options.get("value")
        if value:
            signode += sphinx.addnodes.desc_sig_punctuation("", " = ")
            signode += docutils.nodes.literal(
                text=value, classes=["code", "python"], language="python"
            )
        return result

    directive_cls.handle_signature = handle_signature  # type: ignore


_monkey_patch_pyattribute_handle_signature(sphinx.domains.python.PyAttribute)
_monkey_patch_pyattribute_handle_signature(sphinx.domains.python.PyVariable)



"""Adds synopsis support to the std domain."""

from typing import cast, Optional, List, Iterator, Tuple

import docutils.nodes
import sphinx.application
import sphinx.builders
import sphinx.domains
import sphinx.domains.std
import sphinx.util.nodes
from typing_extensions import Literal

from . import object_description_options
from .. import sphinx_utils


StandardDomain = sphinx.domains.std.StandardDomain
GenericObject = sphinx.domains.std.GenericObject


def _monkey_patch_generic_object_to_support_synopses():

    orig_after_content = GenericObject.after_content

    orig_transform_content = GenericObject.transform_content

    def transform_content(self: GenericObject, contentnode) -> None:
        setattr(self, "contentnode", contentnode)
        orig_transform_content(self, contentnode)

    GenericObject.transform_content = transform_content

    def after_content(self: GenericObject) -> None:
        orig_after_content(self)
        noindex = "noindex" in self.options
        if noindex:
            return
        options = object_description_options.get_object_description_options(
            self.env, self.domain, self.objtype
        )
        generate_synopses = options["generate_synopses"]
        if generate_synopses is None:
            return
        synopsis = sphinx_utils.summarize_element_text(
            getattr(self, "contentnode"), generate_synopses
        )
        std = cast(StandardDomain, self.env.get_domain("std"))
        for name in self.names:
            std.data["synopses"][self.objtype, name] = synopsis

    GenericObject.after_content = after_content

    orig_merge_domaindata = StandardDomain.merge_domaindata

    def merge_domaindata(self, docnames: List[str], otherdata: dict) -> None:
        orig_merge_domaindata(self, docnames, otherdata)
        self.data["synopses"].update(otherdata["synopses"])

    StandardDomain.merge_domaindata = merge_domaindata

    def make_refnode(
        std: StandardDomain,
        builder: sphinx.builders.Builder,
        fromdocname: str,
        docname: str,
        labelid: str,
        contnode: docutils.nodes.Element,
        objtype: str,
        target: str,
    ):
        return sphinx.util.nodes.make_refnode(
            builder,
            fromdocname,
            docname,
            labelid,
            contnode,
            title=object_description_options.format_object_description_tooltip(
                env=cast(sphinx.environment.BuildEnvironment, builder.env),
                options=object_description_options.get_object_description_options(
                    std.env, "std", objtype
                ),
                base_title=target,
                synopsis=std.data["synopses"].get((objtype, target)),
            ),
        )

    def _resolve_obj_xref(
        self: StandardDomain,
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        typ: str,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> Optional[docutils.nodes.Element]:
        objtypes = self.objtypes_for_role(typ) or []
        objtype = None
        for objtype in objtypes:
            result = self.objects.get((objtype, target))
            if result is not None:
                docname, labelid = result
                break
        else:
            return None

        return make_refnode(
            self, builder, fromdocname, docname, labelid, contnode, objtype, target
        )

    StandardDomain._resolve_obj_xref = _resolve_obj_xref

    def resolve_any_xref(
        self: StandardDomain,
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> List[Tuple[str, docutils.nodes.Element]]:
        results: List[Tuple[str, docutils.nodes.Element]] = []
        ltarget = target.lower()  # :ref: lowercases its target automatically
        for role in ("ref", "option"):  # do not try "keyword"
            res = self.resolve_xref(
                env,
                fromdocname,
                builder,
                role,
                ltarget if role == "ref" else target,
                node,
                contnode,
            )
            if res:
                results.append(("std:" + role, res))
        # all others
        for objtype in self.object_types:
            key = (objtype, target)
            if objtype == "term":
                key = (objtype, ltarget)
            if key in self.objects:
                docname, labelid = self.objects[key]
                results.append(
                    (
                        "std:" + self.role_for_objtype(objtype),
                        make_refnode(
                            self,
                            builder,
                            fromdocname,
                            docname,
                            labelid,
                            contnode,
                            objtype,
                            target,
                        ),
                    )
                )
        return results

    StandardDomain.resolve_any_xref = resolve_any_xref

    def get_object_synopses(
        self: StandardDomain,
    ) -> Iterator[Tuple[Tuple[str, str], str]]:
        synopses = self.data["synopses"]
        for (objtype, name), (docname, labelid) in self.objects.items():
            synopsis = synopses.get((objtype, name))
            if not synopsis:
                continue
            yield ((docname, labelid), synopsis)

    StandardDomain.get_object_synopses = get_object_synopses


def setup(app: sphinx.application.Sphinx):
    app.setup_extension("sphinx_immaterial.apidoc.object_description_options")
    sphinx.domains.std.StandardDomain.initial_data[
        "synopses"
    ] = {}  # (type, name) -> synopsis
    _monkey_patch_generic_object_to_support_synopses()

    object_description_options.add_object_description_option(
        app,
        "generate_synopses",
        type_constraint=Optional[Literal["first_paragraph", "first_sentence"]],
        default="first_paragraph",
    )

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Extensions to the sphinx.ext.graphviz module."""

import html
import io
import os
import pathlib
import re
import subprocess
import tempfile
from typing import Optional, Any, Tuple, Dict, Type, Sequence, NamedTuple
import xml.etree.ElementTree as ET

import docutils.nodes
import sphinx.application
import sphinx.ext.graphviz
import sphinx.util.docutils
import sphinx.util.logging
from sphinx.writers.html import HTMLTranslator

from . import external_resource_cache
from . import sphinx_utils

logger = sphinx.util.logging.getLogger(__name__)


def _replace_resolved_xrefs(node: sphinx.ext.graphviz.graphviz, code: str) -> str:
    """Extracts any resolved references, and uses them to replace `xref`
    attributes in the DOT code.
    """
    ref_replacements = {}

    for child in node.children:
        if not isinstance(child, docutils.nodes.container):
            continue
        xref_id = child.get("xref_id")
        if xref_id is None:
            continue
        text = child.astext()
        ref_nodes = list(child.findall(condition=docutils.nodes.reference))

        title = None
        url = None
        target = None
        replacement_text = f"label=<{html.escape(text)}>"
        if ref_nodes:
            ref_node = ref_nodes[-1]
            refuri = ref_node.get("refuri")
            if refuri is not None:
                url = refuri or "#"
            else:
                url = "#" + ref_node["refid"]
            title = ref_node.get("reftitle")
            replacement_text += f' href="{url}"'
            if title is not None:
                replacement_text += f" tooltip=<{html.escape(title)}>"
            target = ref_node.get("target")
            if target is not None:
                replacement_text += f' target="{target}"'

        ref_replacements[xref_id] = replacement_text

    if ref_replacements:
        ref_pattern = "|".join(ref_replacements.keys())
        code = re.sub(ref_pattern, lambda m: ref_replacements[m.group(0)], code)
    return code


class GraphvizConfigInfo(NamedTuple):
    """Adjusted graphviz configuration.

    This extension configures Graphviz to use use the same font (from Google
    Fonts) that is used for other document text.

    Note that when generating SVG output, Graphviz uses the font just to compute
    the size of labels.  The web browser is still responsible for loading the
    font and rendering the text.  Therefore, the font needs to be available both
    to Graphviz and the web browser.

    Since the font already needs to be available to the web browser for other
    document text, there is no added complication there.  But making the font
    available to Graphviz is non-trivial:

    - Graphviz supports several mechanisms for text layout:

      - LibGD, always available on Linux, sometimes on Windows, and allows TTF
        font paths to be specified directly;

      - Pango and/or GDI+ (Windows only), relies on systems-specific font paths
        and does not allow paths to TTF fonts to be specified directly.

      Therefore, LibGD must be used, but unfortunately, depending on how
      Graphviz is built, Pango and/or GDI+ is normally used by default.

    - Graphviz does not provide any command-line options for controlling which
      plugins are loaded.  Instead, plugins can only be loaded through a config
      file.

    Therefore, to ensure LibGD is used for fonts, it is necessary to locate the
    original config file, parse it and remove the section that loads Pango, if
    present, and write it to a new temporary directory.

    """

    orig_config_path: str
    """Path to original config file."""

    new_config: bytes
    """New config content with pango excluded."""


def get_adjusted_graphviz_config(
    app: sphinx.application.Sphinx, dot_command: str
) -> Optional[GraphvizConfigInfo]:
    """Returns the graphviz configuration info for a given `dot_command`.

    The returned config file must be written to a temporary directory with a
    symlink "plugins" to the `orig_config_path`.
    """
    key = "_sphinx_immaterial_graphviz_adjusted_configs"
    configs = getattr(app, key, None)
    if configs is None:
        configs = {}
        setattr(app, key, configs)
    config = configs.get(dot_command, False)
    if config is False:
        config = _make_adjusted_graphviz_config(app, dot_command)
        configs[dot_command] = config
    return config


def _get_orig_config_path(dot_command: str) -> Optional[str]:
    result = subprocess.run(
        [dot_command, "-v"], input="", text=True, capture_output=True, check=True
    )

    m = re.search(
        r"^The plugin configuration file:\s+(.*)$\s+was successfully loaded\.$",
        result.stderr,
        re.MULTILINE,
    )
    if m is None:
        logger.error(
            "Failed to determine graphviz config path from stderr: %r", result.stderr
        )
        return None

    return m.group(1)


def _make_adjusted_graphviz_config(
    app: sphinx.application.Sphinx, dot_command: str
) -> Optional[GraphvizConfigInfo]:
    """Determines the graphviz libdir and generates an adjusted config.

    This is called by `get_adjusted_graphviz_config`.
    """

    orig_config_path = _get_orig_config_path(dot_command)
    if orig_config_path is None:
        return None

    config_content = pathlib.Path(orig_config_path).read_bytes()

    # Strip comments
    config_content = re.sub(b"#[^\n]*", b"", config_content)

    new_config = io.BytesIO()

    prev_index = 0

    def parse_error():
        logger.error(
            "Failed to parse graphviz config file %r, starting at: %r",
            orig_config_path,
            config_content[prev_index:],
        )

    found_gd = False

    # Match plugins
    for m in re.finditer(
        rb"\s*([^\s{}]+)\s+([^\s{}]+)\s*(\{\s*(?:[^{}\s]+\s*\{[^{}]*\}\s*)*\s*\})\s*",
        config_content,
    ):
        if m.start() != prev_index:
            parse_error()
            return None
        prev_index = m.end()
        plugin_path = m.group(1)
        plugin_name = m.group(2)
        plugin_config = m.group(3)
        if plugin_name == b"gd":
            found_gd = True
        else:
            plugin_config = re.sub(rb"\btextlayout\s*\{[^}]*\}", b"", plugin_config)
        if not os.path.isabs(plugin_path):
            plugin_path = os.path.join(b"plugins", plugin_path)
        new_config.write(plugin_path)
        new_config.write(b" ")
        new_config.write(plugin_name)
        new_config.write(b" ")
        new_config.write(plugin_config)

    if prev_index != len(config_content):
        parse_error()
        return None

    if not found_gd:
        if not app.config.graphviz_ignore_incorrect_font_metrics:
            logger.warning(
                "Incorrect font metrics will be used because "
                "graphviz binary %r does not have LibGD support.  This warning is expected on x86_64 Windows "
                "(https://gitlab.com/graphviz/graphviz/-/issues/2267). "
                "Set `graphviz_ignore_incorrect_font_metrics = True` in `conf.py` "
                "to silence this warning.",
                dot_command,
            )
        return None

    return GraphvizConfigInfo(
        orig_config_path=orig_config_path, new_config=new_config.getvalue()
    )


def render_dot_html(
    self: HTMLTranslator,
    node: sphinx.ext.graphviz.graphviz,
    code: str,
    options: dict,
    prefix: str = "graphviz",
    imgcls: str = None,
    alt: Optional[str] = None,
    filename: Optional[str] = None,
) -> Tuple[str, str]:

    theme_options = self.builder.config["html_theme_options"]
    ttf_font_paths = external_resource_cache.get_ttf_font_paths(self.builder.app)
    font = theme_options["font"]["text"]

    ttf_font = ttf_font_paths[(font, "400")]

    code = _replace_resolved_xrefs(node, code)

    var_replacements: Dict[str, str] = {}
    replacement_to_var: Dict[str, str] = {}

    def replace_var(var_text: str) -> str:
        replacement_color = var_replacements.setdefault(
            var_text, "#%06x" % (0x123456 + len(var_replacements))
        )
        replacement_to_var.setdefault(replacement_color, var_text)
        return replacement_color

    def replace_var_in_code(m: re.Match) -> str:
        var_text = m.group(1)
        replacement_color = replace_var(var_text)
        return f'"{replacement_color}"'

    fontcolor = replace_var("var(--md-code-fg-color)")
    fontsize = "12"

    graphviz_dot = options.get("graphviz_dot", self.builder.config.graphviz_dot)
    config_info = get_adjusted_graphviz_config(self.builder.app, graphviz_dot)

    if config_info is None:
        ttf_font = font

    command_line_options = [
        "-Ncolor=" + replace_var("var(--md-graphviz-node-fg-color)"),
        "-Nstyle=solid,filled",
        "-Nfillcolor=" + replace_var("var(--md-graphviz-node-bg-color)"),
        "-Nfontcolor=" + fontcolor,
        "-Nfontsize=" + fontsize,
        "-Nfontname=" + ttf_font,
        "-Ecolor=" + replace_var("var(--md-graphviz-edge-color)"),
        "-Efontcolor=" + fontcolor,
        "-Efontsize=" + fontsize,
        "-Efontname=" + ttf_font,
        "-Gbgcolor=transparent",
        "-Gcolor=" + replace_var("var(--md-graphviz-node-fg-color)"),
        "-Gfontcolor=" + fontcolor,
        "-Gfontsize=" + fontsize,
        "-Gfontname=" + ttf_font,
    ]

    code = re.sub(r'"((?:var|calc)\s*\(.*?\))"', replace_var_in_code, code)

    dot_cmd = [graphviz_dot]
    dot_cmd.extend(command_line_options)
    dot_cmd.extend(self.builder.config.graphviz_dot_args)
    dot_cmd.append("-Tsvg")

    with tempfile.TemporaryDirectory() as tempdir:
        env = os.environ.copy()
        if config_info is not None:
            orig_lib_path = pathlib.Path(config_info.orig_config_path)
            new_lib_dir = pathlib.Path(tempdir, "plugins")
            pathlib.Path(tempdir, orig_lib_path.name).write_bytes(
                config_info.new_config
            )

            env["GVBINDIR"] = tempdir
            new_lib_dir.symlink_to(orig_lib_path.parent, target_is_directory=True)
            cwd = str(orig_lib_path.parent)
        else:
            cwd = None
        dot_result = subprocess.run(
            dot_cmd,
            input=code,
            encoding="utf-8",
            check=False,
            env=env,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        svg_output = dot_result.stdout
        errors = dot_result.stderr.strip()
        if config_info is None:
            # Filter warnings about fonts
            errors = re.sub(r"^.*couldn\'t load font .*$", "", errors, re.MULTILINE)
        errors = errors.strip()
        if errors or dot_result.returncode != 0:
            error_func = logger.warning if dot_result.returncode == 0 else logger.error
            error_func(
                "Error running %r with env %r: %s", dot_cmd, env, errors, location=node
            )
            if dot_result.returncode != 0:
                raise docutils.nodes.SkipNode

    ET.register_namespace("", "http://www.w3.org/2000/svg")
    root = ET.fromstring(svg_output)
    xlink_href_key = "{http://www.w3.org/1999/xlink}href"
    xlink_title_key = "{http://www.w3.org/1999/xlink}title"
    a_tag = "{http://www.w3.org/2000/svg}a"
    text_tag = "{http://www.w3.org/2000/svg}text"

    for element in root.iter():
        style = ""
        attrib = element.attrib
        for attr in ("fill", "stroke"):
            attrib_val = attrib.get(attr)
            if attrib_val is None:
                continue
            var_val = replacement_to_var.get(attrib_val)
            if var_val is not None:
                del attrib[attr]
                style += f"{attr}: {var_val};"
        font_family = attrib.get("font-family")
        if font_family == ttf_font:
            attrib["font-family"] = font
        href = attrib.pop(xlink_href_key, None)
        if href is not None:
            attrib["href"] = href
        title = attrib.pop(xlink_title_key, None)
        if title is not None:
            title_element = ET.Element("title")
            title_element.text = title
            element.append(title_element)
        if element.tag == a_tag:
            for child in element:
                if child.tag == text_tag:
                    child.attrib["within_a"] = "true"
        within_a = attrib.pop("within_a", None)
        if within_a:
            style += "--md-graphviz-hover-color: var(--md-graphviz-a-hover-color);"
        if style:
            attrib["style"] = style

    classes = [imgcls, "graphviz"] + node.get("classes", [])
    imgcls = " ".join(filter(None, classes))
    root.attrib["class"] = (root.attrib.get("class", "") + " " + imgcls).strip()

    base_scale = 0.75

    def convert_width_or_height(s: str):
        assert s.endswith("pt")
        val_pt = float(s[:-2])
        val_px = val_pt / 0.75
        val_rem = val_px / 16 * base_scale
        return f"{val_rem}rem"

    root_style = ""
    for attr in ("width", "height"):
        attrib_val = root.attrib.pop(attr, None)
        if attrib_val is not None:
            root_style += f"{attr}: {convert_width_or_height(attrib_val)};"
    if root_style:
        root.attrib["style"] = root_style

    svg_output = ET.tostring(root, encoding="unicode")

    if alt is None:
        alt = node.get("alt", self.encode(code).strip())
    if "align" in node:
        self.body.append(
            '<div align="%s" class="align-%s">' % (node["align"], node["align"])
        )
    self.body.append(svg_output)
    if "align" in node:
        self.body.append("</div>\n")

    raise docutils.nodes.SkipNode


def _replace_var_refs_with_defaults(code: str) -> str:
    code = re.sub(r'"var\s*\(.*?,\s*(.*)\)"', lambda m: f'"{m.group(1)}"', code)
    return code


def on_build_finished(app: sphinx.application.Sphinx, exc: Optional[Exception]) -> None:
    # Suppress inclusion of the graphviz.css file supplied by
    # `sphinx.ext.graphviz`.  This theme provides its own style rules.
    pass


sphinx.ext.graphviz.on_build_finished = on_build_finished


def _preprocess_graphviz_node(
    directive: sphinx.util.docutils.SphinxDirective,
    node: sphinx.ext.graphviz.graphviz,
    line_offset: int,
) -> None:
    code = node["code"]

    xrefs: Dict[str, Tuple[str, int]] = {}

    def replace_xref(m: re.Match) -> str:
        line_index = code.count("\n", 0, m.start())
        xref_text = m.group(1).replace(r"\"", '"')
        xref_index = len(xrefs)
        xref_id = f"__SPHINX_IMMATERIAL_XREF_{xref_index}__"
        xref_id, _ = xrefs.setdefault(xref_text, (xref_id, line_index - line_offset))
        return xref_id

    code = re.sub(r'\bxref\s*=\s*"((?:[^\\"]*|(?:\\.|"))*)"', replace_xref, code)

    node["code"] = code

    filename = node.get("filename")

    for xref_text, (xref_id, line_index) in xrefs.items():
        container = docutils.nodes.container()
        container["xref_id"] = xref_id

        # Determine source location
        if filename is None:
            source_path, source_offset = directive.content.items[line_index]
        else:
            source_path = os.path.join(directive.env.app.srcdir, filename)
            source_offset = line_index

        nodes = sphinx_utils.parse_rst(
            state=directive.state,
            text=xref_text,
            source_path=source_path,
            source_line=source_offset,
        )
        container += nodes
        node += container


def _monkey_patch_graphviz_directive(
    directive: Type[sphinx.util.docutils.SphinxDirective], line_offset: int
):
    orig_run = directive.run

    def run(
        self: sphinx.util.docutils.SphinxDirective,
    ) -> Sequence[docutils.nodes.Node]:
        nodes = orig_run(self)
        for node in nodes:
            for graphviz_node in node.findall(condition=sphinx.ext.graphviz.graphviz):
                _preprocess_graphviz_node(self, graphviz_node, line_offset)
        return nodes

    directive.run = run  # type: ignore[assignment]


_monkey_patch_graphviz_directive(sphinx.ext.graphviz.Graphviz, line_offset=0)
_monkey_patch_graphviz_directive(sphinx.ext.graphviz.GraphvizSimple, line_offset=2)


def _monkey_patch_render_dot(name: str) -> None:
    orig_render = getattr(sphinx.ext.graphviz, name)

    def render_dot(
        self, node: sphinx.ext.graphviz.graphviz, code: str, options: Dict, **kwargs
    ):
        code = _replace_resolved_xrefs(node, code)
        code = _replace_var_refs_with_defaults(code)
        return orig_render(self, node, code, options, **kwargs)

    setattr(sphinx.ext.graphviz, name, render_dot)


sphinx.ext.graphviz.render_dot_html = render_dot_html
_monkey_patch_render_dot("render_dot_texinfo")
_monkey_patch_render_dot("render_dot_latex")


def setup(app: sphinx.application.Sphinx) -> Dict[str, Any]:
    app.setup_extension("sphinx.ext.graphviz")
    app.add_config_value(
        "graphviz_ignore_incorrect_font_metrics",
        types=(bool,),
        default=False,
        rebuild="env",
    )
    sphinx_utils.remove_css_file(app, "graphviz.css")
    return {"parallel_read_safe": True, "parallel_write_safe": True}



import docutils.nodes
import docutils.statemachine

from sphinx.domains.python import PyObject
import sphinx.util.nodes


def _monkey_patch_python_domain_to_support_titles():
    """Enables support for titles in all Python directive types.

    Normally sphinx only supports titles in `automodule`, but the python_apigen
    extension uses titles to group member summaries.
    """

    orig_before_content = PyObject.before_content

    def before_content(self: PyObject) -> None:
        orig_before_content(self)
        setattr(self, "_saved_content", self.content)
        self.content = docutils.statemachine.StringList()

    orig_transform_content = PyObject.transform_content

    def transform_content(self: PyObject, contentnode: docutils.nodes.Node) -> None:
        sphinx.util.nodes.nested_parse_with_titles(
            self.state,
            getattr(self, "_saved_content"),
            contentnode,
        )
        orig_transform_content(self, contentnode)

    PyObject.before_content = before_content
    PyObject.transform_content = transform_content


_monkey_patch_python_domain_to_support_titles()



"""Python API documentation generation extension.

A separate page is generated for each class/function/member/constant to be
documented.

As with sphinx.ext.autosummary, we have to physically write a separate rST file
to the source tree for each object to document, as an initial preprocesing step,
since that provides the simplest way to get Sphinx to process those pages.  It
is recommended to run the build with the source tree copied to a temporary
directory in order to avoid modifying the real source tree.

Unlike the sphinx.ext.autosummary extension, we use Sphinx Python domain
directives for the "summaries" as well, rather than a plain table, in order to
display the signatures nicely.
"""

import copy
import dataclasses
import hashlib
import importlib
import inspect
import json
import os
import re
from typing import (
    List,
    Tuple,
    Any,
    Optional,
    Type,
    cast,
    Dict,
    NamedTuple,
    Iterator,
    Set,
)

import docutils.nodes
import docutils.parsers.rst.states
import docutils.statemachine
import sphinx
import sphinx.addnodes
import sphinx.application
import sphinx.domains.python
import sphinx.environment
import sphinx.ext.autodoc
import sphinx.ext.autodoc.directive
import sphinx.ext.napoleon.docstring
import sphinx.pycode
import sphinx.util.docstrings
import sphinx.util.docutils
import sphinx.util.inspect
import sphinx.util.logging
import sphinx.util.typing

from .. import object_description_options
from ... import sphinx_utils
from .. import apigen_utils
from ... import default_literal_role

logger = sphinx.util.logging.getLogger(__name__)

_UNCONDITIONALLY_DOCUMENTED_MEMBERS = frozenset(
    [
        "__init__",
        "__class_getitem__",
        "__call__",
        "__getitem__",
        "__setitem__",
    ]
)
"""Special members to include even if they have no docstring."""


class ParsedOverload(NamedTuple):
    """Parsed representation of a single overload.

    For non-function types and non-overloaded functions, this just represents the
    object itself.

    Sphinx does not really support pybind11-style overloaded functions directly.
    It has minimal support functions with multiple signatures, with a single
    docstring.  However, pybind11 produces overloaded functions each with their
    own docstring.  This module adds support for documenting each overload as an
    independent function.

    Additionally, we need a way to identify each overload, for the purpose of
    generating a page name, listing in the table of contents sidebar, and
    cross-referencing.  Sphinx does not have a native solution to this problem
    because it is not designed to support overloads.  Doxygen uses some sort of
    hash as the identifier, but that means links break with even minor changes to
    the signature.

    Instead, we require that a unique id be manually assigned to each overload,
    and specified as:

        Overload:
          XXX

    in the docstring.  Then the overload will be identified as
    `module.Class.function(overload)`, and will be documented using the page name
    `module.Class.function-overload`.  Typically the overload id should be chosen
    to be a parameter name that is unique to the overload.
    """

    doc: Optional[str]
    """Docstring for individual overload.  First line is the signature."""

    overload_id: Optional[str] = None
    """Overload id specified in the docstring.

  If there is just a single overload, will be `None`.  Otherwise, if no overload
  id is specified, a warning is produced and the index of the overload,
  i.e. "1", "2", etc., is used as the id.
  """


def _extract_field(doc: str, field: str) -> Tuple[str, Optional[str]]:
    pattern = f"\n\\s*\n{field}:\\s*\n\\s+([^\n]+)\n"
    m = re.search(pattern, doc)
    if m is None:
        return doc, None
    start, end = m.span()
    return f"{doc[:start]}\n\n{doc[end:]}", m.group(1).strip()


_OVERLOADED_FUNCTION_RE = "^([^(]+)\\([^\n]*\nOverloaded function.\n"


def _parse_overloaded_function_docstring(doc: Optional[str]) -> List[ParsedOverload]:
    """Parses a pybind11 overloaded function docstring.

    If the docstring is not for an overloaded function, just returns the full
    docstring as a single "overload".

    :param doc: Original docstring.
    :returns: List of parsed overloads.
    :raises ValueError: If docstring has unexpected format.
    """

    if doc is None:
        return [ParsedOverload(doc=doc, overload_id=None)]
    m = re.match(_OVERLOADED_FUNCTION_RE, doc)
    if m is None:
        # Non-overloaded function
        doc, overload_id = _extract_field(doc, "Overload")
        return [ParsedOverload(doc=doc, overload_id=overload_id)]

    display_name = m.group(1)
    doc = doc[m.end() :]
    i = 1

    def get_prefix(i: int):
        return "\n%d. %s(" % (i, display_name)

    prefix = get_prefix(i)
    parts: List[ParsedOverload] = []
    while doc:
        if not doc.startswith(prefix):
            raise ValueError(
                "Docstring does not contain %r as expected: %r"
                % (
                    prefix,
                    doc,
                )
            )
        doc = doc[len(prefix) - 1 :]
        nl_index = doc.index("\n")
        part_sig = doc[:nl_index]
        doc = doc[nl_index + 1 :]
        i += 1
        prefix = get_prefix(i)
        end_index = doc.find(prefix)
        if end_index == -1:
            part = doc
            doc = ""
        else:
            part = doc[:end_index]
            doc = doc[end_index:]

        part, overload_id = _extract_field(part, "Overload")
        if overload_id is None:
            overload_id = str(i - 1)

        part_doc_with_sig = f"{display_name}{part_sig}\n{part}"
        parts.append(
            ParsedOverload(
                doc=part_doc_with_sig,
                overload_id=overload_id,
            )
        )
    return parts


def _get_overloads_from_documenter(
    documenter: sphinx.ext.autodoc.Documenter,
) -> List[ParsedOverload]:
    docstring = sphinx.util.inspect.getdoc(
        documenter.object,
        documenter.get_attr,
        documenter.env.config.autodoc_inherit_docstrings,
        documenter.parent,
        documenter.object_name,
    )
    return _parse_overloaded_function_docstring(docstring)


def _has_default_value(node: sphinx.addnodes.desc_parameter):
    for sub_node in node.findall(condition=docutils.nodes.literal):
        if "default_value" in sub_node.get("classes"):
            return True
    return False


def _summarize_signature(
    env: sphinx.environment.BuildEnvironment, node: sphinx.addnodes.desc_signature
) -> None:
    """Shortens a signature line to fit within `wrap_signatures_column_limit."""

    obj_desc = node.parent

    options = object_description_options.get_object_description_options(
        env, obj_desc["domain"], obj_desc["objtype"]
    )
    column_limit = options["wrap_signatures_column_limit"]

    def _must_shorten():
        return len(node.astext()) > column_limit

    parameterlist: Optional[sphinx.addnodes.desc_parameterlist] = None
    for parameterlist in node.findall(condition=sphinx.addnodes.desc_parameterlist):
        break

    if parameterlist is None:
        # Can't shorten a signature without a parameterlist
        return

    # Remove initial `self` parameter
    if parameterlist.children and parameterlist.children[0].astext() == "self":
        del parameterlist.children[0]

    added_ellipsis = False
    for next_parameter_index in range(len(parameterlist.children) - 1, -1, -1):
        if not _must_shorten():
            return

        # First remove type annotation of last parameter, but only if it doesn't
        # have a default value.
        last_parameter = parameterlist.children[next_parameter_index]
        if isinstance(
            last_parameter, sphinx.addnodes.desc_parameter
        ) and not _has_default_value(last_parameter):
            del last_parameter.children[1:]
            if not _must_shorten():
                return

        # Elide last parameter entirely
        del parameterlist.children[next_parameter_index]
        if not added_ellipsis:
            added_ellipsis = True
            ellipsis_node = sphinx.addnodes.desc_sig_punctuation("", "...")
            param = sphinx.addnodes.desc_parameter()
            param += ellipsis_node
            parameterlist += param


class _MemberDocumenterEntry(NamedTuple):
    """Represents a member of some outer scope (module/class) to document."""

    documenter: sphinx.ext.autodoc.Documenter
    is_attr: bool
    name: str
    """Member name within parent, e.g. class member name."""

    full_name: str
    """Full name under which to document the member.

    For example, "modname.ClassName.method".
    """

    parent_canonical_full_name: str

    overload: Optional[ParsedOverload] = None

    is_inherited: bool = False
    """Indicates whether this is an inherited member."""

    subscript: bool = False
    """Whether this is a "subscript" method to be shown with [] instead of ()."""

    @property
    def overload_suffix(self):
        if self.overload and self.overload.overload_id:
            return f"({self.overload.overload_id})"
        return ""

    @property
    def toc_title(self):
        return self.name + self.overload_suffix


class _ApiEntityMemberReference(NamedTuple):
    name: str
    canonical_object_name: str
    parent_canonical_object_name: str
    inherited: bool = False


@dataclasses.dataclass
class _ApiEntity:
    canonical_full_name: str
    objtype: str
    directive: str
    overload_id: str

    @property
    def canonical_object_name(self):
        return self.canonical_full_name + self.overload_suffix

    @property
    def object_name(self):
        return self.documented_full_name + self.overload_suffix

    @property
    def overload_suffix(self) -> str:
        overload_id = self.overload_id
        return f"({overload_id})" if overload_id else ""

    signatures: List[str]
    options: Dict[str, str]
    content: List[str]
    group_name: str
    order: Optional[int]
    subscript: bool

    members: List[_ApiEntityMemberReference]
    """Members of this entity."""

    parents: List[_ApiEntityMemberReference]
    """Parents that reference this entity as a member.

    Inverse of `members`."""

    docname: str = ""

    documented_full_name: str = ""
    documented_name: str = ""
    top_level: bool = False

    base_classes: Optional[List[str]] = None
    """List of base classes, as rST cross references."""


def _is_constructor_name(name: str) -> bool:
    return name in ("__init__", "__new__", "__class_getitem__")


class _ApiData:

    entities: Dict[str, _ApiEntity]
    top_level_groups: Dict[str, List[_ApiEntityMemberReference]]

    def __init__(self):
        self.entities = {}
        self.top_level_groups = {}

    def get_name_for_signature(
        self, entity: _ApiEntity, member: Optional[_ApiEntityMemberReference]
    ) -> str:
        if member is not None:
            assert member.canonical_object_name == entity.canonical_object_name
            # Get name for summary
            if _is_constructor_name(member.name):
                parent_entity = self.entities.get(member.parent_canonical_object_name)
                if parent_entity is not None and parent_entity.objtype == "class":
                    # Display as the parent class name.
                    return parent_entity.documented_name
            # Display as the member name
            return member.name
        full_name = entity.documented_full_name
        if _is_constructor_name(entity.documented_name):
            full_name = full_name[: -len(entity.documented_name) - 1]
        module = entity.options.get("module")
        if module is not None and full_name.startswith(module + "."):
            # Strip module name, since it is specified separately.
            full_name = full_name[len(module) + 1 :]
        return full_name

    def sort_members(
        self, members: List[_ApiEntityMemberReference], alphabetical=False
    ):
        def get_key(member: _ApiEntityMemberReference):
            member_entity = self.entities[member.canonical_object_name]
            order = member_entity.order or 0
            if alphabetical:
                return (order, member.name.lower(), member.name)
            return order

        members.sort(key=get_key)


def _ensure_module_name_in_signature(signode: sphinx.addnodes.desc_signature) -> None:
    """Ensures non-summary objects are documented with the module name.

    Sphinx by default excludes the module name from class members, and does not
    provide an option to override that.  Since we display all objects on separate
    pages, we want to include the module name for clarity.

    :param signode: Signature to modify in place.
    """
    for node in signode.findall(condition=sphinx.addnodes.desc_addname):
        modname = signode.get("module")
        if modname and not node.astext().startswith(modname + "."):
            node.insert(0, docutils.nodes.Text(modname + "."))
        break


def _get_group_name(
    default_groups: List[Tuple[re.Pattern, str]], entity: _ApiEntity
) -> str:
    """Returns a default group name for an entry.

    This is used if the group name is not explicitly specified via "Group:" in the
    docstring.

    :param default_groups: Default group associations.
    :param entity: Entity to document.
    :returns: The group name.
    """
    s = f"{entity.objtype}:{entity.documented_full_name}"
    group = "Public members"
    for pattern, default_group in default_groups:
        if pattern.fullmatch(s) is not None:
            group = default_group
    return group


def _get_order(default_order: List[Tuple[re.Pattern, int]], entity: _ApiEntity) -> int:
    """Returns a default order value for an entry.

    This is used if the order is not explicitly specified via "Order:" in the
    docstring.

    :param default_order: Default order associations.
    :param entity: Entity to document.
    :returns: The order.
    """
    s = f"{entity.objtype}:{entity.documented_full_name}"
    order = 0
    for pattern, order_value in default_order:
        if pattern.fullmatch(s) is not None:
            order = order_value
    return order


def _mark_subscript_parameterlist(signode: sphinx.addnodes.desc_signature) -> None:
    """Modifies an object description to display as a "subscript method".

    A "subscript method" is a property that defines __getitem__ and is intended to
    be treated as a method invoked using [] rather than (), in order to allow
    subscript syntax like ':'.

    :param node: Signature to modify in place.
    """
    for sub_node in signode.findall(condition=sphinx.addnodes.desc_parameterlist):
        sub_node["parens"] = ("[", "]")


def _clean_init_signature(signode: sphinx.addnodes.desc_signature) -> None:
    """Modifies an object description of an __init__ method.

    Removes the return type (always None) and the self parameter (since these
    methods are displayed as the class name, without showing __init__).

    :param node: Signature to modify in place.
    """
    # Remove first parameter.
    for param in signode.findall(condition=sphinx.addnodes.desc_parameter):
        if param.children[0].astext() == "self":
            param.parent.remove(param)
        break

    # Remove return type.
    for node in signode.findall(condition=sphinx.addnodes.desc_returns):
        node.parent.remove(node)


def _clean_class_getitem_signature(signode: sphinx.addnodes.desc_signature) -> None:
    """Modifies an object description of a __class_getitem__ method.

    Removes the `static` prefix since these methods are shown using the class
    name (i.e. as "subscript" constructors).

    :param node: Signature to modify in place.

    """
    # Remove `static` prefix
    for prefix in signode.findall(condition=sphinx.addnodes.desc_annotation):
        prefix.parent.remove(prefix)
        break


def _get_api_data(
    env: sphinx.environment.BuildEnvironment,
) -> _ApiData:
    return getattr(env, "_sphinx_immaterial_python_apigen_data")


def _generate_entity_desc_node(
    env: sphinx.environment.BuildEnvironment,
    entity: _ApiEntity,
    state: docutils.parsers.rst.states.RSTState,
    member: Optional[_ApiEntityMemberReference] = None,
    callback=None,
):
    api_data = _get_api_data(env)

    summary = member is not None
    name = api_data.get_name_for_signature(entity, member)

    def object_description_transform(
        app: sphinx.application.Sphinx,
        domain: str,
        objtype: str,
        contentnode: sphinx.addnodes.desc_content,
    ) -> None:
        env = app.env
        assert env is not None

        obj_desc = contentnode.parent
        assert isinstance(obj_desc, sphinx.addnodes.desc)
        signodes = cast(List[sphinx.addnodes.desc_signature], obj_desc[:-1])
        if not signodes:
            return

        for signode in signodes:
            fullname = signode["fullname"]
            modname = signode["module"]
            object_name = (modname + "." if modname else "") + fullname
            if object_name != entity.object_name:
                # This callback may be invoked for additional members
                # documented within the body of `entry`, but we don't want
                # to transform them here.
                return
            assert isinstance(signode, sphinx.addnodes.desc_signature)
            if entity.subscript:
                _mark_subscript_parameterlist(signode)

            if entity.documented_name in ("__init__", "__new__"):
                _clean_init_signature(signode)
            if entity.documented_name == "__class_getitem__":
                _clean_class_getitem_signature(signode)

            if summary:
                obj_desc["classes"].append("summary")
                assert app.env is not None
                _summarize_signature(app.env, signode)

            base_classes = entity.base_classes
            if base_classes:
                signode += sphinx.addnodes.desc_sig_punctuation("", "(")
                for i, base_class in enumerate(base_classes):
                    base_entity = api_data.entities.get(base_class)
                    if base_entity is not None:
                        base_class = base_entity.object_name
                    if i != 0:
                        signode += sphinx.addnodes.desc_sig_punctuation("", ",")
                        signode += sphinx.addnodes.desc_sig_space()
                    signode += sphinx.domains.python._parse_annotation(base_class, env)
                signode += sphinx.addnodes.desc_sig_punctuation("", ")")

        if callback is not None:
            callback(contentnode)

    listener_id = env.app.connect(
        "object-description-transform", object_description_transform, priority=999
    )
    content = entity.content
    options = dict(entity.options)
    options["nonodeid"] = ""
    options["object-ids"] = json.dumps([entity.object_name] * len(entity.signatures))
    if summary:
        content = _summarize_rst_content(content)
        options["noindex"] = ""
        options.pop("canonical", None)
    else:
        options["canonical"] = entity.canonical_object_name
    try:
        with apigen_utils.save_rst_defaults(env):
            rst_input = docutils.statemachine.StringList()
            sphinx_utils.append_multiline_string_to_stringlist(
                rst_input,
                ".. highlight-push::\n\n"
                + env.config.python_apigen_rst_prolog
                + "\n\n",
                "<python_apigen_rst_prolog>",
                -2,
            )
            sphinx_utils.append_directive_to_stringlist(
                rst_input,
                entity.directive,
                signatures=[name + sig for sig in entity.signatures],
                content="\n".join(content),
                source_path=entity.object_name,
                source_line=0,
                options=options,
            )
            sphinx_utils.append_multiline_string_to_stringlist(
                rst_input,
                "\n\n"
                + env.config.python_apigen_rst_epilog
                + "\n\n.. highlight-pop::\n\n",
                "<python_apigen_rst_epilog>",
                -2,
            )
            nodes = [
                x
                for x in sphinx_utils.parse_rst(
                    state=state,
                    text=rst_input,
                )
                if isinstance(x, sphinx.addnodes.desc)
            ]
    finally:
        env.app.disconnect(listener_id)

    if len(nodes) != 1:
        raise ValueError("Failed to document entity: %r" % (entity.object_name,))
    node = nodes[0]

    return node


def _generate_entity_summary(
    env: sphinx.environment.BuildEnvironment,
    member: _ApiEntityMemberReference,
    state: docutils.parsers.rst.states.RSTState,
    include_in_toc: bool,
) -> sphinx.addnodes.desc:
    api_data = _get_api_data(env)
    entity = api_data.entities[member.canonical_object_name]
    objdesc = _generate_entity_desc_node(
        env=env, entity=entity, member=member, state=state
    )
    for sig_node in cast(List[sphinx.addnodes.desc_signature], objdesc.children[:-1]):
        # Insert a link around the `desc_name` field
        for sub_node in sig_node.findall(condition=sphinx.addnodes.desc_name):
            if include_in_toc:
                sub_node["classes"].append("pseudo-toc-entry")
            xref_node = sphinx.addnodes.pending_xref(
                "",
                sub_node.deepcopy(),
                refdomain="py",
                reftype="obj",
                reftarget=entity.object_name,
                refwarn=True,
                refexplicit=True,
            )
            sub_node.replace_self(xref_node)
            break
        # Only mark first signature as a `pseudo-toc-entry`.
        include_in_toc = False
    return objdesc


def _generate_group_summary(
    env: sphinx.environment.BuildEnvironment,
    members: List[_ApiEntityMemberReference],
    state: docutils.parsers.rst.states.RSTState,
    notoc: Optional[bool] = None,
):

    data = _get_api_data(env)

    nodes: List[docutils.nodes.Node] = []
    toc_entries: List[Tuple[str, str]] = []

    for member in members:
        member_entity = data.entities[member.canonical_object_name]
        include_in_toc = True
        if notoc is True:
            include_in_toc = False
        elif member is not member_entity.parents[0]:
            include_in_toc = False
        node = _generate_entity_summary(
            env=env, member=member, state=state, include_in_toc=include_in_toc
        )
        if node is None:
            continue
        nodes.append(node)

        if include_in_toc:
            toc_entries.append(
                (member.name + member_entity.overload_suffix, member_entity.docname)
            )

    nodes.extend(
        sphinx_utils.make_toctree_node(
            state,
            toc_entries,
            options={"hidden": True},
            source_path=__name__,
        )
    )
    return nodes


def _add_group_summary(
    env: sphinx.environment.BuildEnvironment,
    contentnode: docutils.nodes.Element,
    sections: Dict[str, docutils.nodes.section],
    group_name: str,
    members: List[_ApiEntityMemberReference],
    state: docutils.parsers.rst.states.RSTState,
) -> None:

    group_id = docutils.nodes.make_id(group_name)
    section = sections.get(group_id)
    if section is None:
        section = docutils.nodes.section()
        section["ids"].append(group_id)
        title = docutils.nodes.title("", group_name)
        section += title
        contentnode += section
        sections[group_id] = section

    section.extend(_generate_group_summary(env=env, members=members, state=state))


def _merge_summary_nodes_into(
    env: sphinx.environment.BuildEnvironment,
    entity: _ApiEntity,
    contentnode: docutils.nodes.Element,
    state: docutils.parsers.rst.states.RSTState,
) -> None:
    """Merges the member summary into `contentnode`.

    Members are organized into groups.  The group is either specified explicitly
    by a `Group:` field in the docstring, or determined automatically by
    `_get_group_name`.  If there is an existing section, the member summary is
    appended to it.  Otherwise, a new section is created.

    :param contentnode: The existing container to which the member summaries will be
        added.  If `contentnode` contains sections, those sections correspond to
        group names.
    """

    sections: Dict[str, docutils.nodes.section] = {}
    for section in contentnode.findall(condition=docutils.nodes.section):
        if section["ids"]:
            sections[section["ids"][0]] = section

    # Maps group name to the list of members.
    groups: Dict[str, List[_ApiEntityMemberReference]] = {}

    entities = _get_api_data(env).entities
    for member in entity.members:
        member_entity = entities[member.canonical_object_name]
        groups.setdefault(member_entity.group_name, []).append(member)

    for group_name, members in groups.items():
        _add_group_summary(
            env=env,
            contentnode=contentnode,
            sections=sections,
            group_name=group_name,
            members=members,
            state=state,
        )


class PythonApigenEntityPageDirective(sphinx.util.docutils.SphinxDirective):
    """Documents an entity and summarizes its members, if applicable."""

    has_content = False
    final_argument_whitespace = True
    required_arguments = 1
    optional_arguments = 0

    def run(self) -> List[docutils.nodes.Node]:
        entities = _get_api_data(self.env).entities
        object_name = self.arguments[0]
        entity = entities.get(object_name)
        if entity is None:
            logger.error(
                "Undefined Python entity: %r", object_name, location=self.get_location()
            )
            return []

        objtype = entity.objtype
        objdesc = _generate_entity_desc_node(
            self.env,
            entity,
            state=self.state,
            callback=lambda contentnode: _merge_summary_nodes_into(
                env=self.env,
                entity=entity,
                contentnode=contentnode,
                state=self.state,
            ),
        )
        if objdesc is None:
            return []

        for signode in objdesc.children[:-1]:
            signode = cast(sphinx.addnodes.desc_signature, signode)
            _ensure_module_name_in_signature(signode)

        # Wrap in a section
        section = docutils.nodes.section()
        section["ids"].append("")
        # Sphinx treates the first child of a `section` node as the title,
        # regardless of its type.  We use a comment node to avoid adding a title
        # that would be redundant with the object description.
        section += docutils.nodes.comment("", entity.object_name)
        section += objdesc
        return [section]


class PythonApigenTopLevelGroupDirective(sphinx.util.docutils.SphinxDirective):
    """Summarizes the members of a top-level group."""

    has_content = False
    final_argument_whitespace = True
    required_arguments = 1
    optional_arguments = 0

    option_spec = {
        "notoc": docutils.parsers.rst.directives.flag,
    }

    def run(self) -> List[docutils.nodes.Node]:
        env = self.env
        data = _get_api_data(env)
        top_level_groups = data.top_level_groups

        group_id = docutils.nodes.make_id(self.arguments[0])
        members = data.top_level_groups.get(group_id)
        if members is None:
            logger.warning(
                "No top-level Python API group named: %r, valid groups are: %r",
                group_id,
                list(data.top_level_groups.keys()),
                location=self.state_machine.get_source_and_line(self.lineno),
            )
            return []
        return _generate_group_summary(
            env=env, members=members, state=self.state, notoc="notoc" in self.options
        )


class PythonApigenEntitySummaryDirective(sphinx.util.docutils.SphinxDirective):
    """Generates a summary/link to a Python entity."""

    has_content = False
    final_argument_whitespace = True
    required_arguments = 1
    optional_arguments = 0

    option_spec = {
        "notoc": docutils.parsers.rst.directives.flag,
    }

    def run(self) -> List[docutils.nodes.Node]:
        env = self.env
        data = _get_api_data(env)
        object_name = self.arguments[0]
        entity = data.entities.get(object_name)
        if entity is None:
            logger.warning(
                "No Python entity named: %r",
                object_name,
                location=self.state_machine.get_source_and_line(self.lineno),
            )
            return []
        return _generate_group_summary(
            env=env,
            members=[entity.parents[0]],
            state=self.state,
            notoc="notoc" in self.options,
        )


class _FakeBridge(sphinx.ext.autodoc.directive.DocumenterBridge):
    def __init__(
        self,
        env: sphinx.environment.BuildEnvironment,
        tab_width: int,
        extra_options: dict,
    ) -> None:
        settings = docutils.parsers.rst.states.Struct(tab_width=tab_width)
        document = docutils.parsers.rst.states.Struct(settings=settings)
        state = docutils.parsers.rst.states.Struct(document=document)
        options = sphinx.ext.autodoc.Options()
        options["undoc-members"] = True
        options["class-doc-from"] = "class"
        options.update(extra_options)
        super().__init__(
            env=env,
            reporter=sphinx.util.docutils.NullReporter(),
            options=options,
            lineno=0,
            state=state,
        )


_EXCLUDED_SPECIAL_MEMBERS = frozenset(
    [
        "__module__",
        "__abstractmethods__",
        "__dict__",
        "__weakref__",
        "__class__",
        "__base__",
        # Exclude pickling members since they are never documented.
        "__getstate__",
        "__setstate__",
    ]
)


def _create_documenter(
    env: sphinx.environment.BuildEnvironment,
    documenter_cls: Type[sphinx.ext.autodoc.Documenter],
    name: str,
    tab_width: int = 8,
) -> sphinx.ext.autodoc.Documenter:
    """Creates a documenter for the given full object name.

    Since we are using the documenter independent of any autodoc directive, we use
    a `_FakeBridge` as the documenter bridge, similar to the strategy used by
    `sphinx.ext.autosummary`.

    :param env: Sphinx build environment.
    :param documenter_cls: Documenter class to use.
    :param name: Full object name, e.g. `my_module.MyClass.my_function`.
    :param tab_width: Tab width setting to use when parsing docstrings.
    :returns: The documenter object.
    """
    extra_options = {}
    if documenter_cls.objtype == "class":
        extra_options["special-members"] = sphinx.ext.autodoc.ALL
    bridge = _FakeBridge(env, tab_width=tab_width, extra_options=extra_options)
    documenter = documenter_cls(bridge, name)
    assert documenter.parse_name()
    assert documenter.import_object()
    try:
        documenter.analyzer = sphinx.pycode.ModuleAnalyzer.for_module(
            documenter.get_real_modname()
        )
        # parse right now, to get PycodeErrors on parsing (results will
        # be cached anyway)
        documenter.analyzer.find_attr_docs()
    except sphinx.pycode.PycodeError:
        # no source file -- e.g. for builtin and C modules
        documenter.analyzer = None  # type: ignore[assignment]
    return documenter


def _get_member_documenter(
    parent: sphinx.ext.autodoc.Documenter,
    member_name: str,
    member_value: Any,
    is_attr: bool,
) -> Optional[sphinx.ext.autodoc.Documenter]:
    """Creates a documenter for the given member.

    :param parent: Parent documenter.
    :param member_name: Name of the member.
    :param member_value: Value of the member.
    :param is_attr: Whether the member is an attribute.
    :returns: The documenter object.
    """
    classes = [
        cls
        for cls in parent.documenters.values()
        if cls.can_document_member(member_value, member_name, is_attr, parent)
    ]
    if not classes:
        return None
    # prefer the documenter with the highest priority
    classes.sort(key=lambda cls: cls.priority)
    full_mname = parent.modname + "::" + ".".join(parent.objpath + [member_name])
    documenter = _create_documenter(
        env=parent.env,
        documenter_cls=classes[-1],
        name=full_mname,
        tab_width=parent.directive.state.document.settings.tab_width,
    )
    return documenter


def _include_member(member_name: str, member_value: Any, is_attr: bool) -> bool:
    """Determines whether a member should be documented.

    :param member_name: Name of the member.
    :param member_value: Value of the member.
    :param is_attr: Whether the member is an attribute.
    :returns: True if the member should be documented.
    """
    del is_attr
    if member_name == "__init__":
        doc = getattr(member_value, "__doc__", None)
        if isinstance(doc, str) and doc.startswith("Initialize self. "):
            return False
    elif member_name in ("__hash__", "__iter__"):
        if member_value is None:
            return False
    return True


def _get_subscript_method(
    parent_documenter: sphinx.ext.autodoc.Documenter, entry: _MemberDocumenterEntry
) -> Any:
    """Checks for a property that defines a subscript method.

    A subscript method is a property like `Class.vindex` where `fget` has a return
    type of `Class._Vindex`, which is a class type.

    :param parent_documenter: Parent documenter for `entry`.
    :param entry: Entry to check.
    :returns: The type object (e.g. `Class._Vindex`) representing the subscript
        method, or None if `entry` does not define a subscript method.
    """
    if not isinstance(entry.documenter, sphinx.ext.autodoc.PropertyDocumenter):
        return None
    retann = entry.documenter.retann
    if not retann:
        return None

    config = entry.documenter.config
    pattern = config.python_apigen_subscript_method_types
    match = pattern.fullmatch(retann)
    if not match:
        return None

    # Attempt to import value
    try:
        mem_documenter = _create_documenter(
            env=entry.documenter.env,
            documenter_cls=sphinx.ext.autodoc.ClassDocumenter,
            name=retann,
        )
        mem = mem_documenter.object
    except ImportError:
        return None
    if not mem:
        return None
    getitem = getattr(mem, "__getitem__", None)
    if getitem is None:
        return None

    return mem


def _transform_member(
    parent_documenter: sphinx.ext.autodoc.Documenter, entry: _MemberDocumenterEntry
) -> Iterator[_MemberDocumenterEntry]:
    """Converts an individual member into a sequence of members to document.

    :param parent_documenter: The parent documenter.
    :param entry: The original entry to document.  For most entries we simply yield the
        entry unmodified.  For entries that correspond to subscript methods,
        though, we yield the __getitem__ member (and __setitem__, if applicable)
        separately.
    :returns: Iterator over modified entries to document.
    """
    if entry.name == "__class_getitem__":
        entry = entry._replace(subscript=True)

    mem = _get_subscript_method(parent_documenter, entry)
    if mem is None:
        yield entry
        return
    retann = entry.documenter.retann

    for suffix in ("__getitem__", "__setitem__"):
        method = getattr(mem, suffix, None)
        if method is None:
            continue
        import_name = f"{retann}.{suffix}"
        if import_name.startswith(entry.documenter.modname + "."):
            import_name = (
                entry.documenter.modname
                + "::"
                + import_name[len(entry.documenter.modname) + 1 :]
            )
        new_documenter = _create_documenter(
            env=parent_documenter.env,
            documenter_cls=sphinx.ext.autodoc.MethodDocumenter,
            name=import_name,
            tab_width=parent_documenter.directive.state.document.settings.tab_width,
        )
        if suffix != "__getitem__":
            new_member_name = f"{entry.name}.{suffix}"
            full_name = f"{entry.full_name}.{suffix}"
            subscript = False
        else:
            new_member_name = f"{entry.name}"
            full_name = entry.full_name
            subscript = True

        yield _MemberDocumenterEntry(
            documenter=new_documenter,
            name=new_member_name,
            is_attr=False,
            full_name=full_name,
            parent_canonical_full_name=entry.parent_canonical_full_name,
            subscript=subscript,
        )


def _prepare_documenter_docstring(entry: _MemberDocumenterEntry) -> None:
    """Initializes `entry.documenter` with the correct docstring.

    This overrides the docstring based on `entry.overload` if applicable.

    This must be called before using `entry.documenter`.

    :param entry: Entry to prepare.
    """

    if entry.overload and (
        entry.overload.overload_id is not None
        or
        # For methods, we don't need `ModuleAnalyzer`, so it is safe to always
        # override the normal mechanism of obtaining the docstring.
        # Additionally, for `__init__` and `__new__` we need to specify the
        # docstring explicitly to work around
        # https://github.com/sphinx-doc/sphinx/pull/9518.
        isinstance(entry.documenter, sphinx.ext.autodoc.MethodDocumenter)
    ):
        # Force autodoc to use the overload-specific signature.  autodoc already
        # has an internal mechanism for overriding the docstrings based on the
        # `_new_docstrings` member.
        tab_width = entry.documenter.directive.state.document.settings.tab_width
        setattr(
            entry.documenter,
            "_new_docstrings",
            [
                sphinx.util.docstrings.prepare_docstring(
                    entry.overload.doc or "", tabsize=tab_width
                )
            ],
        )
    else:
        # Force autodoc to obtain the docstring through its normal mechanism,
        # which includes the "ModuleAnalyzer" for reading docstrings of
        # variables/attributes that are only contained in the source code.
        setattr(entry.documenter, "_new_docstrings", None)

    # Workaround for https://github.com/sphinx-doc/sphinx/pull/9518
    orig_get_doc = entry.documenter.get_doc

    def get_doc(*args, **kwargs) -> List[List[str]]:
        doc_strings = getattr(entry.documenter, "_new_docstrings", None)
        if doc_strings is not None:
            return doc_strings
        return orig_get_doc(*args, **kwargs)  # type: ignore

    entry.documenter.get_doc = get_doc  # type: ignore[assignment]


def _is_conditionally_documented_entry(entry: _MemberDocumenterEntry):
    if entry.name in _UNCONDITIONALLY_DOCUMENTED_MEMBERS:
        return False
    return sphinx.ext.autodoc.special_member_re.match(entry.name)


def _get_member_overloads(
    entry: _MemberDocumenterEntry,
) -> Iterator[_MemberDocumenterEntry]:
    """Returns the list of overloads for a given entry."""

    if entry.name in _EXCLUDED_SPECIAL_MEMBERS:
        return

    overloads = _get_overloads_from_documenter(entry.documenter)
    for overload in overloads:
        # Shallow copy the documenter.  Certain methods on the documenter mutate it,
        # and we don't want those mutations to affect other overloads.
        documenter_copy = copy.copy(entry.documenter)
        documenter_copy.options = documenter_copy.options.copy()
        new_entry = entry._replace(
            overload=overload,
            documenter=documenter_copy,
        )
        if _is_conditionally_documented_entry(new_entry):
            # Only document this entry if it has a docstring.
            _prepare_documenter_docstring(new_entry)
            new_entry.documenter.format_signature()
            doc = new_entry.documenter.get_doc()
            if not doc:
                continue
            if not any(x for x in doc):
                # No docstring, skip.
                continue

            new_entry = entry._replace(
                overload=overload, documenter=copy.copy(entry.documenter)
            )

        yield new_entry


def _get_documenter_direct_members(
    documenter: sphinx.ext.autodoc.Documenter,
    canonical_full_name: str,
) -> Iterator[_MemberDocumenterEntry]:
    """Returns the sequence of direct members to document.

    The order is mostly determined by the definition order.

    This excludes inherited members.

    :param documenter: Documenter for which to obtain members.
    :returns: Iterator over members to document.
    """
    if not isinstance(
        documenter,
        (sphinx.ext.autodoc.ClassDocumenter, sphinx.ext.autodoc.ModuleDocumenter),
    ):
        # Only classes and modules have members.
        return

    members_check_module, members = documenter.get_object_members(want_all=True)
    del members_check_module
    if members:
        try:
            # get_object_members does not preserve definition order, but __dict__ does
            # in Python 3.6 and later.
            member_dict = sphinx.util.inspect.safe_getattr(
                documenter.object, "__dict__"
            )
            member_order = {k: i for i, k in enumerate(member_dict.keys())}
            members.sort(key=lambda entry: member_order.get(entry[0], float("inf")))
        except AttributeError:
            pass
    filtered_members = [
        x
        for x in documenter.filter_members(members, want_all=True)
        if _include_member(*x)
    ]
    for member_name, member_value, is_attr in filtered_members:
        member_documenter = _get_member_documenter(
            parent=documenter,
            member_name=member_name,
            member_value=member_value,
            is_attr=is_attr,
        )
        if member_documenter is None:
            continue
        full_name = f"{documenter.fullname}.{member_name}"
        entry = _MemberDocumenterEntry(
            cast(sphinx.ext.autodoc.Documenter, member_documenter),
            is_attr,
            parent_canonical_full_name=canonical_full_name,
            name=member_name,
            full_name=full_name,
        )
        for transformed_entry in _transform_member(documenter, entry):
            yield from _get_member_overloads(transformed_entry)


def _get_documenter_members(
    documenter: sphinx.ext.autodoc.Documenter,
    canonical_full_name: str,
) -> Iterator[_MemberDocumenterEntry]:
    """Returns the sequence of members to document, including inherited members.

    :param documenter: Parent documenter for which to find members.
    :returns: Iterator over members to document.
    """
    seen_members: Set[str] = set()

    def _get_unseen_members(
        members: Iterator[_MemberDocumenterEntry], is_inherited: bool
    ) -> Iterator[_MemberDocumenterEntry]:
        for member in members:
            overload_name = member.toc_title
            if overload_name in seen_members:
                continue
            seen_members.add(overload_name)
            yield member._replace(is_inherited=is_inherited)

    yield from _get_unseen_members(
        _get_documenter_direct_members(
            documenter, canonical_full_name=canonical_full_name
        ),
        is_inherited=False,
    )

    if documenter.objtype != "class":
        return

    for cls in inspect.getmro(documenter.object):
        if cls is documenter.object:
            continue
        if cls.__module__ in ("builtins", "pybind11_builtins"):
            continue
        class_name = f"{cls.__module__}::{cls.__qualname__}"
        parent_canonical_full_name = f"{cls.__module__}.{cls.__qualname__}"
        try:
            superclass_documenter = _create_documenter(
                env=documenter.env,
                documenter_cls=sphinx.ext.autodoc.ClassDocumenter,
                name=class_name,
                tab_width=documenter.directive.state.document.settings.tab_width,
            )
            yield from _get_unseen_members(
                _get_documenter_direct_members(
                    superclass_documenter,
                    canonical_full_name=parent_canonical_full_name,
                ),
                is_inherited=True,
            )
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(
                "Cannot obtain documenter for base class %r of %r: %r",
                cls,
                documenter.fullname,
                e,
            )


class SplitAutodocRstOutput(NamedTuple):
    directive: str
    options: Dict[str, str]
    content: List[str]
    group_name: Optional[str]
    order: Optional[int]


def _split_autodoc_rst_output(
    rst_strings: docutils.statemachine.StringList,
) -> SplitAutodocRstOutput:
    m = re.fullmatch(r"\.\. ([^:]+:[^:]+):: (.*)", rst_strings[1], re.DOTALL)
    assert m is not None, repr(rst_strings[1])
    directive = m.group(1)
    signatures = [m.group(2)]
    signature_prefix = " " * (6 + len(directive))
    i = 2
    while i < len(rst_strings):
        line = rst_strings[i]
        if line.startswith(signature_prefix):
            signatures.append(line[len(signature_prefix) :])
            i += 1
        else:
            break
    options: Dict[str, str] = {}
    while i < len(rst_strings):
        line = rst_strings[i]
        m = re.fullmatch(r"   :([^:]+):(.*)", line, re.DOTALL)
        if m is None:
            break
        options[m.group(1)] = m.group(2).strip()
        i += 1
    assert i < len(rst_strings)
    assert rst_strings[i] == ""

    i += 1

    while i < len(rst_strings) and not rst_strings[i].strip():
        i += 1

    content_start_line = i

    def extract_field(name: str) -> Tuple[Optional[str], Optional[Tuple[str, int]]]:
        field_prefix = f"   :{name}:"
        for i in range(content_start_line, len(rst_strings)):
            line = rst_strings[i]
            if not line.startswith(field_prefix):
                continue
            value = line[len(field_prefix) :].strip()
            location = rst_strings.items[i]
            del rst_strings[i]
            return value, location
        return None, None

    group_name, group_location = extract_field("group")

    order_str, order_location = extract_field("order")
    order = None
    if order_str is not None:
        try:
            order = int(order_str)
        except ValueError:
            logger.error("Invalid order value: %r", order_str, location=order_location)

    # Strip 3 spaces of indent.
    content = [line[3:] for line in rst_strings.data[content_start_line:]]

    return SplitAutodocRstOutput(
        directive=directive,
        options=options,
        content=content,
        group_name=group_name,
        order=order,
    )


def _summarize_rst_content(content: List[str]) -> List[str]:
    i = 0
    # Skip over blank lines before start of directive content
    while i < len(content) and not content[i].strip():
        i += 1
    # Skip over first paragraph
    while i < len(content) and content[i].strip():
        i += 1

    return content[:i]


class _ApiEntityCollector:
    def __init__(
        self,
        entities: Dict[str, _ApiEntity],
    ):
        self.entities = entities

    def collect_entity_recursively(
        self,
        entry: _MemberDocumenterEntry,
    ) -> str:
        canonical_full_name = None
        if isinstance(entry.documenter, sphinx.ext.autodoc.ClassDocumenter):
            canonical_full_name = entry.documenter.get_canonical_fullname()
        if canonical_full_name is None:
            canonical_full_name = f"{entry.parent_canonical_full_name}.{entry.name}"

        canonical_object_name = canonical_full_name + entry.overload_suffix
        existing_entity = self.entities.get(canonical_object_name)
        if existing_entity is not None:
            return canonical_object_name

        if (
            entry.overload
            and entry.overload.overload_id
            and re.fullmatch("[0-9]+", entry.overload.overload_id)
        ):
            logger.warning("Unspecified overload id: %s", canonical_object_name)

        rst_strings = docutils.statemachine.StringList()
        entry.documenter.directive.result = rst_strings
        _prepare_documenter_docstring(entry)
        # Prevent autodoc from also documenting members, since this extension does
        # that separately.
        def document_members(*args, **kwargs):
            return

        entry.documenter.document_members = document_members  # type: ignore[assignment]
        entry.documenter.generate()

        base_classes: Optional[List[str]] = None

        if isinstance(entry.documenter, sphinx.ext.autodoc.ClassDocumenter):
            # By default (unless the `autodoc_class_signature` config option is
            # set to `"separated"`), autodoc will include the `__init__`
            # parameters in the signature.  Since that convention does not work
            # well with this extension, we just bypass that here.
            signatures = [""]

            if entry.documenter.config.python_apigen_show_base_classes:
                obj = entry.documenter.object
                bases = sphinx.util.inspect.getorigbases(obj)
                if bases is None:
                    bases = getattr(obj, "__bases__", None)
                if bases:
                    base_list = list(bases)
                    entry.documenter.env.events.emit(
                        "autodoc-process-bases",
                        entry.documenter.fullname,
                        obj,
                        entry.documenter.options,
                        base_list,
                    )
                    base_classes = [
                        sphinx.util.typing.stringify(base)
                        for base in base_list
                        if base is not object
                    ]
        else:
            signatures = entry.documenter.format_signature().split("\n")

        split_result = _split_autodoc_rst_output(rst_strings)

        split_result.options.pop("module", None)

        overload_id: Optional[str] = None
        if entry.overload is not None:
            overload_id = entry.overload.overload_id

        entity = _ApiEntity(
            documented_full_name="",
            canonical_full_name=canonical_full_name,
            objtype=entry.documenter.objtype,
            group_name=split_result.group_name or "",
            order=split_result.order,
            directive=split_result.directive,
            signatures=signatures,
            options=split_result.options,
            content=split_result.content,
            members=[],
            parents=[],
            subscript=entry.subscript,
            overload_id=overload_id or "",
            base_classes=base_classes,
        )

        self.entities[canonical_object_name] = entity

        entity.members = self.collect_documenter_members(
            entry.documenter,
            canonical_object_name=canonical_object_name,
        )

        return canonical_object_name

    def collect_documenter_members(
        self,
        documenter: sphinx.ext.autodoc.Documenter,
        canonical_object_name: str,
    ) -> List[_ApiEntityMemberReference]:
        members: List[_ApiEntityMemberReference] = []

        for entry in _get_documenter_members(
            documenter, canonical_full_name=canonical_object_name
        ):
            member_canonical_object_name = self.collect_entity_recursively(entry)
            member = _ApiEntityMemberReference(
                name=entry.name,
                parent_canonical_object_name=canonical_object_name,
                canonical_object_name=member_canonical_object_name,
                inherited=entry.is_inherited,
            )
            members.append(member)
            child = self.entities[member_canonical_object_name]
            child.parents.append(member)

        return members


def _get_base_docname(output_prefixes: Dict[str, str], full_name: str) -> str:
    end_idx = len(full_name)
    while True:
        output_path = output_prefixes.get(full_name[:end_idx])
        if output_path is not None:
            return output_path + full_name[end_idx + 1 :]
        new_end_idx = full_name.rfind(".", 0, end_idx)
        if new_end_idx == -1:
            raise ValueError(
                f"Could not find output prefix for {full_name!r} in {output_prefixes!r}"
            )
        end_idx = new_end_idx


def _get_docname(
    output_prefixes: Dict[str, str],
    documented_full_name: str,
    overload_id: str,
    case_insensitive_filesystem: bool,
):

    docname = _get_base_docname(output_prefixes, documented_full_name)
    if overload_id:
        docname += f"-{overload_id}"

    return apigen_utils.make_unique_docname(docname, case_insensitive_filesystem)


def _assign_documented_full_names(
    entities: Dict[str, _ApiEntity],
    default_groups: List[Tuple[re.Pattern, str]],
    default_order: List[Tuple[re.Pattern, int]],
    output_prefixes: Dict[str, str],
    case_insensitive_filesystem: bool,
) -> None:
    """Determines the full name under which each entity will be documented."""

    def get_documented_full_name(entity: _ApiEntity) -> str:
        documented_full_name = entity.documented_full_name
        if documented_full_name:
            # Name already assigned
            return documented_full_name

        parents = entity.parents
        assert len(parents) > 0

        def parent_sort_key(parent_ref: _ApiEntityMemberReference):
            canonical_name_from_parent = (
                parent_ref.parent_canonical_object_name + "." + parent_ref.name
            )
            canonical = canonical_name_from_parent == entity.canonical_full_name
            inherited = parent_ref.inherited
            return (canonical is False, inherited, canonical_name_from_parent)

        parents.sort(key=parent_sort_key)

        parent_ref = parents[0]
        parent_entity = entities.get(parent_ref.parent_canonical_object_name)
        if parent_entity is None:
            # Parent is a module.
            parent_documented_name = parent_ref.parent_canonical_object_name
            entity.top_level = True
            entity.options["module"] = parent_ref.parent_canonical_object_name
        else:
            parent_documented_name = get_documented_full_name(parent_entity)
            entity.options["module"] = parent_entity.options["module"]
        documented_full_name = parent_documented_name + "." + parent_ref.name
        entity.documented_full_name = documented_full_name
        entity.documented_name = parent_ref.name

        # Assign default group name.
        if not entity.group_name:
            entity.group_name = _get_group_name(default_groups, entity)

        if entity.order is None:
            entity.order = _get_order(default_order, entity)

        entity.docname = _get_docname(
            output_prefixes,
            documented_full_name,
            entity.overload_id,
            case_insensitive_filesystem,
        )

        return documented_full_name

    for entity in list(entities.values()):
        get_documented_full_name(entity)
        entities[entity.object_name] = entity


def _builder_inited(app: sphinx.application.Sphinx) -> None:
    """Generates the rST files for API members."""

    env = app.env
    assert env is not None

    data = _ApiData()

    setattr(env, "_sphinx_immaterial_python_apigen_data", data)

    apigen_modules = app.config.python_apigen_modules
    if not apigen_modules:
        return

    for module_name, output_path in apigen_modules.items():
        try:
            importlib.import_module(module_name)
        except ImportError:
            logger.warning(
                "Failed to import module %s specified in python_apigen_modules",
                module_name,
                exc_info=True,
            )
            continue

        documenter = _create_documenter(
            env=env,
            documenter_cls=sphinx.ext.autodoc.ModuleDocumenter,
            name=module_name,
        )
        _ApiEntityCollector(entities=data.entities,).collect_documenter_members(
            documenter=documenter,
            canonical_object_name=module_name,
        )

    writer = apigen_utils.GeneratedDocumentWriter(
        app=app,
        case_insensitive_filesystem=app.config.python_apigen_case_insensitive_filesystem,
        output_prefixes=list(apigen_modules.values()),
        generator_module=__name__,
    )

    writer.prepare_output_directories()

    default_groups = [
        (re.compile(pattern), group)
        for pattern, group in app.config.python_apigen_default_groups
    ]

    default_order = [
        (re.compile(pattern), order)
        for pattern, order in app.config.python_apigen_default_order
    ]

    _assign_documented_full_names(
        entities=data.entities,
        default_groups=default_groups,
        default_order=default_order,
        output_prefixes=apigen_modules,
        case_insensitive_filesystem=writer.case_insensitive_filesystem,
    )

    alphabetical = app.config.python_apigen_order_tiebreaker == "alphabetical"

    def get_pages():
        for object_name, entity in data.entities.items():
            if object_name != entity.object_name:
                # Alias
                continue

            data.sort_members(entity.members, alphabetical=alphabetical)

            if entity.top_level:
                group_id = docutils.nodes.make_id(entity.group_name)
                data.top_level_groups.setdefault(group_id, []).append(entity.parents[0])

            content = sphinx_utils.format_directive(
                "python-apigen-entity-page",
                entity.object_name,
            )
            yield entity.docname, entity.object_name, content

    writer.write_files(get_pages())

    for members in data.top_level_groups.values():
        data.sort_members(members, alphabetical=alphabetical)


def _monkey_patch_napoleon_to_add_group_field():
    """Adds support to sphinx.ext.napoleon for the "Group" and "Order" fields.

    This field is used by this module to organize members into groups.
    """
    orig_load_custom_sections = (
        sphinx.ext.napoleon.docstring.GoogleDocstring._load_custom_sections
    )  # pylint: disable=protected-access

    def parse_section(
        self: sphinx.ext.napoleon.docstring.GoogleDocstring, section: str
    ) -> List[str]:
        lines = self._strip_empty(
            self._consume_to_next_section()
        )  # pylint: disable=protected-access
        lines = self._dedent(lines)  # pylint: disable=protected-access
        name = section.lower()
        if len(lines) != 1:
            raise ValueError(f"Expected exactly one {name} in {section} section")
        return [f":{name}: " + lines[0], ""]

    def load_custom_sections(
        self: sphinx.ext.napoleon.docstring.GoogleDocstring,
    ) -> None:
        orig_load_custom_sections(self)
        self._sections["group"] = lambda section: parse_section(
            self, section
        )  # pylint: disable=protected-access
        self._sections["order"] = lambda section: parse_section(
            self, section
        )  # pylint: disable=protected-access

    sphinx.ext.napoleon.docstring.GoogleDocstring._load_custom_sections = (
        load_custom_sections  # pylint: disable=protected-access
    )


def _config_inited(
    app: sphinx.application.Sphinx, config: sphinx.config.Config
) -> None:
    if isinstance(config.python_apigen_subscript_method_types, str):
        setattr(
            config,
            "python_apigen_subscript_method_types",
            re.compile(config.python_apigen_subscript_method_types),
        )


def setup(app: sphinx.application.Sphinx):
    """Initializes the extension."""
    _monkey_patch_napoleon_to_add_group_field()
    app.connect("builder-inited", _builder_inited)
    app.connect("config-inited", _config_inited)
    app.setup_extension("sphinx.ext.autodoc")
    app.add_directive("python-apigen-entity-page", PythonApigenEntityPageDirective)
    app.add_directive("python-apigen-group", PythonApigenTopLevelGroupDirective)
    app.add_directive(
        "python-apigen-entity-summary", PythonApigenEntitySummaryDirective
    )
    app.add_config_value(
        "python_apigen_modules", types=(Dict[str, str],), default={}, rebuild="env"
    )
    app.add_config_value(
        "python_apigen_default_groups",
        types=(List[Tuple[str, str]],),
        default=[(".*", "Public members"), ("class:.*", "Classes")],
        rebuild="env",
    )
    app.add_config_value(
        "python_apigen_default_order",
        types=(List[Tuple[str, int]],),
        default=[],
        rebuild="env",
    )
    app.add_config_value(
        "python_apigen_order_tiebreaker",
        types=sphinx.config.ENUM("definition_order", "alphabetical"),
        default="definition_order",
        rebuild="env",
    )
    app.add_config_value(
        "python_apigen_subscript_method_types",
        default=r".*\._[^.]*",
        types=(re.Pattern,),
        rebuild="env",
    )
    app.add_config_value(
        "python_apigen_case_insensitive_filesystem",
        default=None,
        types=(bool, type(None)),
        rebuild="env",
    )
    app.add_config_value(
        "python_apigen_show_base_classes",
        default=True,
        types=(bool,),
        rebuild="env",
    )
    app.add_config_value(
        "python_apigen_rst_prolog", types=(str,), default="", rebuild="env"
    )
    app.add_config_value(
        "python_apigen_rst_epilog", types=(str,), default="", rebuild="env"
    )
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Modifies HTML translation of sections.

Sphinx normally writes sections with a section heading as:

    <div id="identifier" class="section"><hN>...</hN>...</div>

but that is incompatible with the way scroll-margin-top and the
`:target` selector are used in the mkdocs-material CSS.  For
compatibility with mkdocs-material, we strip the outer `<div>` and
instead add the `id` to the inner `<hN>` element.

That is accomplished by overriding `visit_section` and
`depart_section` not to add the `<div>` and `</div>` tags, and also
modifying `visit_title` to insert the `id`.
"""

from typing import Callable

import docutils.nodes

from . import html_translator_mixin


@html_translator_mixin.override
def visit_section(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: docutils.nodes.section,
    super_func: html_translator_mixin.BaseVisitCallback[docutils.nodes.section],
) -> None:
    self.section_level += 1


@html_translator_mixin.override
def depart_section(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: docutils.nodes.section,
    super_func: html_translator_mixin.BaseVisitCallback[docutils.nodes.section],
) -> None:
    self.section_level -= 1


@html_translator_mixin.override
def visit_title(
    self: html_translator_mixin.HTMLTranslatorMixin,
    node: docutils.nodes.title,
    super_func: html_translator_mixin.BaseVisitCallback[docutils.nodes.title],
) -> None:
    if isinstance(node.parent, docutils.nodes.section):
        if node.parent.get("ids") and not node.get("ids"):
            node["ids"] = node.parent.get("ids")
            super_func(self, node)
            del node["ids"]
            return
    super_func(self, node)



class Bravo:
    """
    Class Bravo docstring.
    """

    def my_method_Bravo(self) -> int:
        """
        This is my_method_Bravo.

        This is paragraph two.
        """
        pass

    @property
    def my_property_Bravo(self) -> str:
        """
        This is my_property_Bravo.

        This is paragraph two.
        """
        pass



"""Common utilities needed by the language-specific apigen modules."""

import contextlib
import glob
import hashlib
import os
import pathlib
import secrets
from typing import Optional, List, Dict, Iterator, Tuple, TypeVar, Callable

import docutils.nodes
import sphinx.application
import sphinx.environment
import sphinx.util.logging

from .. import default_literal_role
from .. import sphinx_utils

logger = sphinx.util.logging.getLogger(__name__)

T = TypeVar("T")


def _is_case_insensitive_filesystem(path: str, initial_comment: str) -> bool:
    suffix = secrets.token_hex(16)
    temp_path = path + suffix + "a.rst"
    try:
        pathlib.Path(temp_path).write_text(initial_comment, encoding="utf-8")
        return os.path.exists(path + suffix + "A.rst")
    finally:
        os.remove(temp_path)


def _is_generated_file(rst_path: str, initial_comment: str) -> bool:
    try:
        if os.path.islink(rst_path) or not os.path.isfile(rst_path):
            return False
        content = pathlib.Path(rst_path).read_text(encoding="utf-8")
        return content.startswith(initial_comment)
    except:  # pylint: disable=bare-except
        return False


def make_unique_docname(orig_docname: str, case_insensitive_filesystem: bool) -> str:
    if not case_insensitive_filesystem:
        return orig_docname
    name_hash = hashlib.sha256(
        os.path.basename(orig_docname).encode("utf-8")
    ).hexdigest()[:8]
    return f"{orig_docname}-{name_hash}"


class GeneratedDocumentWriter:
    def __init__(
        self,
        app: sphinx.application.Sphinx,
        case_insensitive_filesystem: Optional[bool],
        output_prefixes: List[str],
        generator_module: str,
    ):
        self.app = app
        self._case_insensitive_filesystem = case_insensitive_filesystem
        self.initial_comment = f"..\n  DO NOT EDIT. GENERATED by {generator_module}.\n"
        self.output_prefixes = output_prefixes
        self.all_pages: Dict[str, str] = {}

    def prepare_output_directories(self) -> None:
        seen_output_dirs = set()
        for output_prefix in self.output_prefixes:
            output_dir = os.path.dirname(os.path.join(self.app.srcdir, output_prefix))
            if output_dir in seen_output_dirs:
                continue
            seen_output_dirs.add(output_dir)
            os.makedirs(output_dir, exist_ok=True)
            if self._case_insensitive_filesystem is None:
                if _is_case_insensitive_filesystem(
                    os.path.join(self.app.srcdir, output_prefix), self.initial_comment
                ):
                    self._case_insensitive_filesystem = True
        if self._case_insensitive_filesystem is None:
            self._case_insensitive_filesystem = False

    @property
    def case_insensitive_filesystem(self) -> bool:
        value = self._case_insensitive_filesystem
        assert value is not None
        return value

    def clear_existing_generated_files(self):
        srcdir = self.app.srcdir
        for output_prefix in self.output_prefixes:
            glob_pattern = os.path.join(srcdir, output_prefix)
            for p in glob.glob(
                os.path.join(srcdir, output_prefix + "*.rst"), recursive=True
            ):
                if not _is_generated_file(p, self.initial_comment):
                    continue
                try:
                    os.remove(p)
                except OSError as e:
                    logger.warning("Failed to remove stale generated file %r: %s", p, e)

    def write_file(self, docname: str, object_name: str, entity_content: str):

        rst_path = docname + ".rst"
        if rst_path in self.all_pages:
            logger.error(
                "Both %r and %r map to generated path %r",
                self.all_pages[rst_path],
                object_name,
                rst_path,
            )
            return

        self.all_pages[rst_path] = object_name

        content = self.initial_comment
        # Suppress "Edit this page" link since the page is generated.
        content += "\n\n:hide-edit-link:\n\n"
        content += entity_content
        rst_path = os.path.join(self.app.srcdir, docname + ".rst")
        if os.path.exists(rst_path):
            logger.error(
                "Generated documentation page for %r would overwrite existing source file %r",
                object_name,
                rst_path,
            )
            return
        pathlib.Path(rst_path).write_text(content, encoding="utf-8")

    def write_files(self, entities: Iterator[Tuple[str, str, str]]):
        self.clear_existing_generated_files()
        for docname, object_name, entity_content in entities:
            self.write_file(docname, object_name, entity_content)


def collect_sections(
    node: docutils.nodes.Element,
) -> Dict[str, docutils.nodes.section]:
    sections: Dict[str, docutils.nodes.section] = {}
    for target in node.findall(condition=docutils.nodes.target):
        ids = target["ids"]
        next_node: docutils.nodes.Node = target.next_node(ascend=True)
        if not isinstance(next_node, docutils.nodes.section):
            continue
        for section_id in ids:
            sections[section_id] = next_node

    for section in node.findall(condition=docutils.nodes.section):
        ids = section["ids"]
        if not ids:
            continue
        for section_id in ids:
            sections[section_id] = section
    return sections


def merge_groups_into(
    parent: docutils.nodes.Element,
    groups: Dict[str, T],
    insert_group: Callable[[T, docutils.nodes.section], None],
    group_id_prefix: str = "",
):

    sections = collect_sections(parent)

    for group_name, group_data in groups.items():
        group_id = group_id_prefix + docutils.nodes.make_id(group_name)
        section = sections.get(group_id)
        if section is None:
            section = docutils.nodes.section()
            section["ids"].append(group_id)
            title = docutils.nodes.title("", group_name)
            section += title
            parent += section
            sections[group_id] = section

        target_start_index = len(section.children)
        while target_start_index > 0 and isinstance(
            section[target_start_index - 1], docutils.nodes.target
        ):
            target_start_index -= 1

        final_targets = section[target_start_index:]
        del section[target_start_index:]
        try:
            insert_group(group_data, section)
        finally:
            section.extend(final_targets)


@contextlib.contextmanager
def save_rst_defaults(env: sphinx.environment.BuildEnvironment):
    with default_literal_role.save_default_literal_role(), sphinx_utils.save_default_role(
        env
    ):
        yield



"""Fixes for the Python domain."""

import json
import re
from typing import (
    cast,
    Sequence,
    Tuple,
    List,
    Dict,
    Type,
    Optional,
    Any,
    Iterator,
    Union,
)

import docutils.nodes
import docutils.parsers.rst.states
import sphinx
import sphinx.addnodes
import sphinx.application
import sphinx.builders
import sphinx.domains
import sphinx.domains.python
import sphinx.environment
import sphinx.errors
import sphinx.ext.napoleon
import sphinx.util.logging
import sphinx.util.nodes

from ... import sphinx_utils
from . import autodoc_property_type

PythonDomain = sphinx.domains.python.PythonDomain
PyTypedField = sphinx.domains.python.PyTypedField
PyObject = sphinx.domains.python.PyObject


logger = sphinx.util.logging.getLogger(__name__)


def setup(app: sphinx.application.Sphinx):
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from typing import List
from docutils import nodes

try:
    from sphinxcontrib.details.directive import (  # pytype: disable=import-error
        DetailsDirective,
    )

    IS_INSTALLED = True
except ImportError:
    IS_INSTALLED = False


def monkey_patch_details_run():
    """Patch the details directive to respect the class option.
    This solution is a temporary fix pending response from
    https://github.com/tk0miya/sphinxcontrib-details-directive/issues/4
    """

    def run(self) -> List[nodes.container]:
        admonition = nodes.container(
            "",
            classes=self.options.get("class", []) + self.options.get("classes", []),
            opened="open" in self.options,
            type="details",
        )
        textnodes, messages = self.state.inline_text(self.arguments[0], self.lineno)
        admonition += nodes.paragraph(self.arguments[0], "", *textnodes)
        admonition += messages
        self.state.nested_parse(self.content, self.content_offset, admonition)
        self.add_name(admonition)
        return [admonition]

    DetailsDirective.run = run


if IS_INSTALLED:
    monkey_patch_details_run()



from typing import Optional, List

import docutils.nodes
import sphinx.domains.python
import sphinx.environment


def ensure_wrapped_in_desc_type(
    nodes: List[docutils.nodes.Node],
) -> List[docutils.nodes.Node]:
    if len(nodes) != 1 or not isinstance(nodes[0], sphinx.addnodes.desc_type):
        nodes = [sphinx.addnodes.desc_type("", "", *nodes)]
    return nodes


def _monkey_patch_python_parse_annotation():
    """Ensures that type annotations in signatures are wrapped in `desc_type`.

    This allows them to be distinguished from parameter names in CSS rules.
    """
    orig_parse_annotation = sphinx.domains.python._parse_annotation

    def parse_annotation(
        annotation: str, env: Optional[sphinx.environment.BuildEnvironment] = None
    ) -> List[docutils.nodes.Node]:
        return ensure_wrapped_in_desc_type(orig_parse_annotation(annotation, env))

    sphinx.domains.python._parse_annotation = parse_annotation


_monkey_patch_python_parse_annotation()
